#!/usr/bin/env python
# cython: language_level=3
#
#  utils.py
"""
General utility functions.

.. versionchanged:: 1.0.0

	* Removed ``tuple2str`` and ``list2string``.
	  Use :func:`domdf_python_tools.utils.list2str` instead.
	* Removed ``as_text`` and ``word_join``.
	  Import from :mod:`domdf_python_tools.words` instead.
	* Removed ``splitLen``.
	  Use :func:`domdf_python_tools.iterative.split_len` instead.

.. versionchanged:: 2.0.0

	:func:`~domdf_python_tools.iterative.chunks`,
	:func:`~domdf_python_tools.iterative.permutations`,
	:func:`~domdf_python_tools.iterative.split_len`,
	:func:`~domdf_python_tools.iterative.Len`, and
	:func:`~domdf_python_tools.iterative.double_chain`
	moved to :func:`domdf_python_tools.iterative`.

.. versionchanged:: 2.3.0

	Removed :func:`domdf_python_tools.utils.deprecated`.
	Use the new `deprecation-alias <https://pypi.org/project/deprecation-alias/>`_ package instead.

"""
#
#  Copyright © 2018-2022 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  as_text from https://stackoverflow.com/a/40935194
# 		Copyright © 2016 User3759685
# 		Available under the MIT License
#
#  strtobool based on the "distutils" module from CPython.
#  Some docstrings based on the Python documentation.
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#

# stdlib
import contextlib
import inspect
import json
import re
import sys
from io import StringIO
from math import log10
from pprint import pformat
from types import MethodType
from typing import (
		IO,
		TYPE_CHECKING,
		Any,
		Callable,
		Dict,
		Iterable,
		Iterator,
		List,
		Optional,
		Pattern,
		Set,
		Tuple,
		TypeVar,
		Union,
		overload
		)

# this package
import domdf_python_tools.words
from domdf_python_tools.typing import HasHead, String, SupportsLessThan

if TYPE_CHECKING or domdf_python_tools.__docs:  # pragma: no cover
	# 3rd party
	from pandas import DataFrame, Series

	Series.__module__ = "pandas"
	DataFrame.__module__ = "pandas"

_T = TypeVar("_T")

SupportsLessThanT = TypeVar("SupportsLessThanT", bound=SupportsLessThan)

__all__ = [
		"pyversion",
		"SPACE_PLACEHOLDER",
		"cmp",
		"list2str",
		"printr",
		"printt",
		"stderr_writer",
		"printe",
		"str2tuple",
		"strtobool",
		"enquote_value",
		"posargs2kwargs",
		"convert_indents",
		"etc",
		"head",
		"magnitude",
		"trim_precision",
		"double_repr_string",
		"redirect_output",
		"divide",
		"redivide",
		"unique_sorted",
		"replace_nonprinting",
		]

#: The current major python version.
pyversion: int = int(sys.version_info.major)  # Python Version

#: The ``␣`` character.
SPACE_PLACEHOLDER = '␣'


def cmp(x, y) -> int:
	"""
	Implementation of ``cmp`` for Python 3.

	Compare the two objects x and y and return an integer according to the outcome.

	The return value is negative if ``x < y``, zero if ``x == y`` and strictly positive if ``x > y``.
	"""

	return int((x > y) - (x < y))


def list2str(the_list: Iterable[Any], sep: str = ',') -> str:
	"""
	Convert an iterable, such as a list, to a comma separated string.

	:param the_list: The iterable to convert to a string.
	:param sep: Separator to use for the string.

	:return: Comma separated string
	"""

	return sep.join([str(x) for x in the_list])


def printr(
		obj: Any,
		*values: object,
		sep: Optional[str] = ' ',
		end: Optional[str] = '\n',
		file: Optional[IO] = None,
		flush: bool = False,
		) -> None:
	r"""
	Print the :func:`repr` of an object.

	If no objects are given, :func:`~.printr` will just write ``end``.

	:param obj:
	:param \*values: Additional values to print. These are printed verbatim.
	:param sep: The separator between values.
	:param end: The final value to print.
		Setting to ``''`` will leave the insertion point at the end of the printed text.
	:param file: The file to write to.
		If not present or :py:obj:`None`, :py:obj:`sys.stdout` will be used.
	:no-default file:
	:param flush: If :py:obj:`True` the stream is forcibly flushed after printing.
	"""

	print(repr(obj), *values, sep=sep, end=end, file=file, flush=flush)


def printt(
		obj: Any,
		*values: object,
		sep: Optional[str] = ' ',
		end: Optional[str] = '\n',
		file: Optional[IO] = None,
		flush: bool = False,
		) -> None:
	r"""
	Print the type of an object.

	If no objects are given, :func:`~.printt` will just write ``end``.

	:param obj:
	:param \*values: Additional values to print. These are printed verbatim.
	:param sep: The separator between values.
	:param end: The final value to print.
		Setting to ``''`` will leave the insertion point at the end of the printed text.
	:param file: The file to write to.
		If not present or :py:obj:`None`, :py:obj:`sys.stdout` will be used.
	:no-default file:
	:param flush: If :py:obj:`True` the stream is forcibly flushed after printing.
	"""

	print(type(obj), *values, sep=sep, end=end, file=file, flush=flush)


def stderr_writer(
		*values: object,
		sep: Optional[str] = ' ',
		end: Optional[str] = '\n',
		) -> None:
	r"""
	Print ``*values`` to :py:obj:`sys.stderr`, separated by ``sep`` and followed by ``end``.

	:py:obj:`sys.stdout` is flushed before printing, and :py:obj:`sys.stderr` is flushed afterwards.

	If no objects are given, :func:`~.stderr_writer` will just write ``end``.

	:param \*values:
	:param sep: The separator between values.
	:param end: The final value to print.
		Setting to ``''`` will leave the insertion point at the end of the printed text.

	:rtype:

	.. versionchanged:: 3.0.0

		The only permitted keyword arguments are ``sep`` and ``end``.
		Previous versions allowed other keywords arguments supported by :func:`print` but they had no effect.
	"""

	sys.stdout.flush()
	print(*values, sep=sep, end=end, file=sys.stderr, flush=True)
	sys.stderr.flush()


#: Alias of :func:`~.stderr_writer`
printe = stderr_writer


def str2tuple(input_string: str, sep: str = ',') -> Tuple[int, ...]:
	"""
	Convert a comma-separated string of integers into a tuple.

	.. latex:vspace:: -10px
	.. important:: The input string must represent a comma-separated series of integers.
	.. TODO:: Allow custom types, not just :class:`int` (making :class:`int` the default)
	.. latex:vspace:: -20px

	:param input_string: The string to be converted into a tuple
	:param sep: The separator in the string.
	"""

	return tuple(int(x) for x in input_string.split(sep))


def strtobool(val: Union[str, int]) -> bool:
	"""
	Convert a string representation of truth to :py:obj:`True` or :py:obj:`False`.

	If val is an integer then its boolean representation is returned. If val is a boolean it is returned as-is.

	:py:obj:`True` values are ``'y'``, ``'yes'``, ``'t'``, ``'true'``, ``'on'``, ``'1'``, and ``1``.

	:py:obj:`False` values are ``'n'``, ``'no'``, ``'f'``, ``'false'``, ``'off'``, ``'0'``, and ``0``.

	:raises: :py:exc:`ValueError` if ``val`` is anything else.
	"""

	if isinstance(val, int):
		return bool(val)

	val = val.lower()
	if val in {'y', "yes", 't', "true", "on", '1'}:
		return True
	elif val in {'n', "no", 'f', "false", "off", '0'}:
		return False
	else:
		raise ValueError(f"invalid truth value {val!r}")


def enquote_value(value: Any) -> Union[str, bool, float]:
	"""
	Adds single quotes (``'``) to the given value, suitable for use in a templating system such as Jinja2.

	:class:`Floats <float>`, :class:`integers <int>`, :class:`booleans <bool>`, :py:obj:`None`,
	and the strings ``'True'``, ``'False'`` and ``'None'`` are returned as-is.

	:param value: The value to enquote
	"""

	if value in {"True", "False", "None", True, False, None}:
		return value
	elif isinstance(value, (int, float)):
		return value
	elif isinstance(value, str):
		return repr(value)
	else:
		return f"'{value}'"


def posargs2kwargs(
		args: Iterable[Any],
		posarg_names: Union[Iterable[str], Callable],
		kwargs: Optional[Dict[str, Any]] = None,
		) -> Dict[str, Any]:
	"""
	Convert the positional args in ``args`` to kwargs, based on the relative order of ``args`` and ``posarg_names``.

	.. important:: Python 3.8's Positional-Only Parameters (:pep:`570`) are not supported.

	.. versionadded:: 0.4.10

	:param args: List of positional arguments provided to a function.
	:param posarg_names: Either a list of positional argument names for the function, or the function object.
	:param kwargs: Optional mapping of keyword argument names to values.
		The arguments will be added to this dictionary if provided.
	:default kwargs: ``{}``

	:return: Dictionary mapping argument names to values.

	.. versionchanged:: 2.8.0

		The "self" argument for bound methods is ignored.
		For unbound methods (which are just functions) the behaviour is unchanged.
	"""

	if kwargs is None:
		kwargs = {}

	self_arg = None

	if isinstance(posarg_names, MethodType):
		self_arg, *posarg_names = inspect.getfullargspec(posarg_names).args
	elif callable(posarg_names):
		posarg_names = inspect.getfullargspec(posarg_names).args

	for name, arg_value in zip(posarg_names, args):
		if name in kwargs:
			if isinstance(posarg_names, MethodType):
				raise TypeError(f"{posarg_names.__name__}(): got multiple values for argument '{name}'")
			else:
				raise TypeError(f"got multiple values for argument '{name}'")

	kwargs.update(zip(posarg_names, args))

	if self_arg is not None and self_arg in kwargs:
		del kwargs[self_arg]

	# TODO: positional only arguments

	return kwargs


def convert_indents(text: str, tab_width: int = 4, from_: str = '\t', to: str = ' ') -> str:
	r"""
	Convert indentation at the start of lines in ``text`` from tabs to spaces.

	:param text: The text to convert indents in.
	:param tab_width: The number of spaces per tab.
	:param from\_: The indent to convert from.
	:param to: The indent to convert to.
	"""

	output = []
	tab = to * tab_width
	from_size = len(from_)

	for line in text.splitlines():
		indent_count = 0

		while line.startswith(from_):
			indent_count += 1
			line = line[from_size:]

		output.append(f"{tab * indent_count}{line}")

	return '\n'.join(output)


class _Etcetera(str):

	__slots__ = ()

	def __new__(cls):
		return str.__new__(cls, "...")

	def __repr__(self) -> str:
		return str(self)


etc = _Etcetera()
"""
Object that provides an ellipsis string

.. versionadded:: 0.8.0
"""


def head(obj: Union[Tuple, List, "DataFrame", "Series", String, HasHead], n: int = 10) -> Optional[str]:
	"""
	Returns the head of the given object.

	.. versionadded:: 0.8.0

	:param obj:
	:param n: Show the first ``n`` items of ``obj``.

	.. seealso::

		* :func:`textwrap.shorten`, which truncates a string to fit within a given number of characters.
		* :func:`itertools.islice`, which returns the first ``n`` elements from an iterator.
	"""

	if isinstance(obj, tuple) and hasattr(obj, "_fields"):
		# Likely a namedtuple
		if len(obj) <= n:
			return repr(obj)
		else:
			head_of_namedtuple = {k: v for k, v in zip(obj._fields[:n], obj[:n])}  # type: ignore
			repr_fmt = '(' + ", ".join(f"{k}={v!r}" for k, v in head_of_namedtuple.items()) + f", {etc})"
			return obj.__class__.__name__ + repr_fmt

	elif isinstance(obj, (list, tuple)):
		if len(obj) > n:
			return pformat(obj.__class__((*obj[:n], etc)))
		else:
			return pformat(obj)

	elif isinstance(obj, HasHead):
		return obj.head(n).to_string()

	elif len(obj) <= n:  # type: ignore
		return str(obj)

	else:
		return str(obj[:n]) + etc  # type: ignore


def magnitude(x: float) -> int:
	"""
	Returns the magnitude of the given value.

	* For negative numbers the absolute magnitude is returned.
	* For decimal numbers below ``1`` the magnitude will be negative.

	.. versionadded:: 2.0.0

	:param x: Numerical value to find the magnitude of.
	"""

	if x > 0.0:
		return int(log10(x))
	elif x < 0.0:
		return int(log10(abs(x)))
	else:
		return 0


def trim_precision(value: float, precision: int = 4) -> float:
	"""
	Trim the precision of the given floating point value.

	For example, if you have the value `170.10000000000002` but really only care about it being
	``\u2248 179.1``:

	.. code-block:: python

		>>> trim_precision(170.10000000000002, 2)
		170.1
		>>> type(trim_precision(170.10000000000002, 2))
		<class 'float'>

	.. versionadded:: 2.0.0

	:param value:
	:param precision: The number of decimal places to leave in the output.
	"""

	return float(format(value, f"0.{precision}f"))


def double_repr_string(string: str) -> str:
	"""
	Like :func:`repr(str) <repr>`, but tries to use double quotes instead.

	.. versionadded:: 2.5.0

	:param string:
	"""

	# figure out which quote to use; double is preferred
	if '"' in string and "'" not in string:
		return repr(string)
	else:
		return json.dumps(string, ensure_ascii=False)


@contextlib.contextmanager
def redirect_output(combine: bool = False) -> Iterator[Tuple[StringIO, StringIO]]:
	"""
	Context manager to redirect stdout and stderr to two :class:`io.StringIO` objects.

	These are assigned (as a :class:`tuple`) to the target the :keyword:`as` expression.

	Example:

	.. code-block:: python

		with redirect_output() as (stdout, stderr):
			...

	.. versionadded:: 2.6.0

	:param combine: If :py:obj:`True` ``stderr`` is combined with ``stdout``.
	"""

	if combine:
		stdout = stderr = StringIO()
	else:
		stdout = StringIO()
		stderr = StringIO()

	with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
		yield stdout, stderr


def divide(string: str, sep: str) -> Tuple[str, str]:
	"""
	Divide a string into two parts, about the given string.

	.. versionadded:: 2.7.0

	:param string:
	:param sep: The separator to split at.
	"""

	if sep not in string:
		raise ValueError(f"{sep!r} not in {string!r}")

	parts = string.split(sep, 1)
	return tuple(parts)  # type: ignore


def redivide(string: str, pat: Union[str, Pattern]) -> Tuple[str, str]:
	"""
	Divide a string into two parts, splitting on the given regular expression.

	.. versionadded:: 2.7.0

	:param string:
	:param pat:

	:rtype:

	.. latex:clearpage::
	"""

	if isinstance(pat, str):
		pat = re.compile(pat)

	if not pat.search(string):
		raise ValueError(f"{pat!r} has no matches in {string!r}")

	parts = pat.split(string, 1)
	return tuple(parts)  # type: ignore


@overload
def unique_sorted(
		elements: Iterable[SupportsLessThanT],
		*,
		key: None = ...,
		reverse: bool = ...,
		) -> List[SupportsLessThanT]: ...


@overload
def unique_sorted(
		elements: Iterable[_T],
		*,
		key: Callable[[_T], SupportsLessThan],
		reverse: bool = ...,
		) -> List[_T]: ...


def unique_sorted(
		elements: Iterable,
		*,
		key: Optional[Callable] = None,
		reverse: bool = False,
		) -> List:
	"""
	Returns an ordered list of unique items from ``elements``.

	.. versionadded:: 3.0.0

	:param elements:
	:param key: A function of one argument used to extract a comparison key from each item when sorting.
		For example, :meth:`key=str.lower <str.lower>`.
		The default value is :py:obj:`None`, which will compare the elements directly.
	:param reverse: If :py:obj:`True` the list elements are sorted as if each comparison were reversed.

	.. seealso:: :class:`set` and :func:`sorted`
	"""

	return sorted(set(elements), key=key, reverse=reverse)


def replace_nonprinting(string: str, exclude: Optional[Set[int]] = None) -> str:
	"""
	Replace nonprinting (control) characters in ``string`` with ``^`` and ``M-`` notation.

	.. versionadded:: 3.3.0

	:param string:
	:param exclude: A set of codepoints to exclude.

	:rtype:

	.. seealso:: :wikipedia:`C0 and C1 control codes` on Wikipedia
	"""

	# https://stackoverflow.com/a/44952259

	if exclude is None:
		exclude = set()

	translation_map = {}

	for codepoint in range(32):
		if codepoint not in exclude:
			translation_map[codepoint] = f"^{chr(64 + codepoint)}"

	if 127 not in exclude:
		translation_map[127] = "^?"

	for codepoint in range(128, 256):
		if codepoint not in exclude:
			translation_map[codepoint] = f"M+{chr(codepoint-64)}"

	return string.translate(translation_map)



#  !/usr/bin/env python
#
#  utils.py
"""
Tools for working with pagesizes.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on reportlab.lib.pagesizes and reportlab.lib.units
#    www.reportlab.co.uk
#    Copyright ReportLab Europe Ltd. 2000-2017
#    Copyright (c) 2000-2018, ReportLab Inc.
#    All rights reserved.
#    Licensed under the BSD License
#
#  Includes data from en.wikipedia.org.
#  Licensed under the Creative Commons Attribution-ShareAlike License
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from typing import Sequence, Tuple, Union, overload

# this package
from domdf_python_tools.typing import AnyNumber

# this package
from .units import Unit, cm, inch, mm, pc, pt, um

# from .units import Unit

__all__ = ["convert_from", "parse_measurement"]


@overload
def convert_from(value: Sequence[AnyNumber], from_: AnyNumber) -> Tuple[float, ...]: ...


@overload
def convert_from(value: AnyNumber, from_: AnyNumber) -> float: ...


def convert_from(
		value: Union[Sequence[AnyNumber], AnyNumber],
		from_: AnyNumber,
		) -> Union[float, Tuple[float, ...]]:
	r"""
	Convert ``value`` to point from the unit specified in ``from_``.

	:param value:
	:param from\_: The unit to convert from, specified as a number of points.
	"""

	if isinstance(value, Sequence):
		return _sequence_convert_from(value, from_)
	else:
		return _sequence_convert_from((value, ), from_)[0]


def _sequence_convert_from(seq: Sequence[AnyNumber], from_: AnyNumber) -> Tuple[float, ...]:
	if isinstance(from_, Unit):
		from_ = from_._in_pt
	else:
		from_ = float(from_)

	return tuple(float(x) * from_ for x in seq)


_measurement_re = re.compile(r"(\d*\.?\d+) *([A-Za-zμµ\"']*)")


def parse_measurement(measurement: str) -> Union[float, Tuple[float, ...]]:
	"""
	Parse the given measurement.

	:param measurement:
	"""

	# TODO: docstring
	all_matches = _measurement_re.findall(measurement)

	if len(all_matches) > 1:
		raise ValueError("Too many measurements")
	elif len(all_matches) == 0:
		raise ValueError("Unable to parse measurement")

	val, unit = all_matches[0]

	if '' in {val, unit}:
		raise ValueError("Unable to parse measurement")

	val = float(val)

	if unit == "mm":
		return val * mm
	elif unit == "cm":
		return val * cm
	elif unit in {"um", "μm", "µm"}:  # second is mu, third is micro
		return val * um
	elif unit == "pt":
		return val * pt
	elif unit in {"inch", "in", '"'}:
		return val * inch
	elif unit in ("pc", "pica"):
		return val * pc

	raise ValueError("Unknown unit")



#  !/usr/bin/env python
#
#  dates.py
"""
Utilities for working with dates and times.

.. extras-require:: dates
	:pyproject:


**Data:**

.. autosummary::

	~domdf_python_tools.dates.months
	~domdf_python_tools.dates.month_full_names
	~domdf_python_tools.dates.month_short_names

"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Parts of the docstrings based on the Python 3.8.2 Documentation
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  calc_easter from https://code.activestate.com/recipes/576517-calculate-easter-western-given-a-year/
#  Copyright © 2008 Martin Diers
#  Licensed under the MIT License
#

# stdlib
import datetime
import sys
import time
import typing
from collections import OrderedDict
from types import ModuleType
from typing import Optional, Union

__all__ = [
		"current_tzinfo",
		"set_timezone",
		"utc_timestamp_to_datetime",
		"months",
		"parse_month",
		"get_month_number",
		"check_date",
		"calc_easter",
		"month_short_names",
		"month_full_names",
		"is_bst",
		]


def current_tzinfo() -> Optional[datetime.tzinfo]:
	"""
	Returns a tzinfo object for the current timezone.
	"""

	return datetime.datetime.now().astimezone().tzinfo  # pragma: no cover (hard to test)


#
# def datetime_to_utc_timestamp(datetime, current_tzinfo=None):
# 	"""
# 	Convert a :class:`datetime.datetime` object to seconds since UNIX epoch, in UTC time
#
# 	:param datetime:
# 	:type datetime: :class:`datetime.datetime`
# 	:param current_tzinfo: A tzinfo object representing the current timezone.
# 		If None it will be inferred.
# 	:type current_tzinfo: :class:`datetime.tzinfo`
#
# 	:return: Timestamp in UTC timezone
# 	:rtype: float
# 	"""
#
# 	return datetime.astimezone(current_tzinfo).timestamp()
#


def set_timezone(obj: datetime.datetime, tzinfo: datetime.tzinfo) -> datetime.datetime:
	"""
	Sets the timezone / tzinfo of the given :class:`datetime.datetime` object.
	This will not convert the time (i.e. the hours will stay the same).
	Use :meth:`datetime.datetime.astimezone` to accomplish that.

	:param obj:
	:param tzinfo:
	"""

	return obj.replace(tzinfo=tzinfo)


def utc_timestamp_to_datetime(
		utc_timestamp: Union[float, int],
		output_tz: Optional[datetime.tzinfo] = None,
		) -> datetime.datetime:
	"""
	Convert UTC timestamp (seconds from UNIX epoch) to a :class:`datetime.datetime` object.

	If ``output_tz`` is :py:obj:`None` the timestamp is converted to the platform’s local date and time,
	and the local timezone is inferred and set for the object.

	If ``output_tz`` is not :py:obj:`None`, it must be an instance of a :class:`datetime.tzinfo` subclass,
	and the timestamp is converted to ``output_tz``’s time zone.


	:param utc_timestamp: The timestamp to convert to a datetime object
	:param output_tz: The timezone to output the datetime object for.
		If :py:obj:`None` it will be inferred.

	:return: The timestamp as a datetime object.

	:raises OverflowError: if the timestamp is out of the range
		of values supported by the platform C localtime() or gmtime() functions,
		and OSError on localtime() or gmtime() failure. It’s common for this to
		be restricted to years in 1970 through 2038.
	"""

	new_datetime = datetime.datetime.fromtimestamp(utc_timestamp, output_tz)
	return new_datetime.astimezone(output_tz)


if sys.version_info <= (3, 7, 2):  # pragma: no cover (py37+)
	MonthsType = OrderedDict
else:  # pragma: no cover (<py37)
	MonthsType = typing.OrderedDict[str, str]  # type: ignore  # noqa: TYP006

#: Mapping of 3-character shortcodes to full month names.
months: MonthsType = OrderedDict(
		Jan="January",
		Feb="February",
		Mar="March",
		Apr="April",
		May="May",
		Jun="June",
		Jul="July",
		Aug="August",
		Sep="September",
		Oct="October",
		Nov="November",
		Dec="December",
		)

month_short_names = tuple(months.keys())
"""
List of the short names for months in the Gregorian calendar.

.. versionadded:: 2.0.0
"""

month_full_names = tuple(months.values())
"""
List of the full names for months in the Gregorian calendar.

.. versionadded:: 2.0.0
"""


def parse_month(month: Union[str, int]) -> str:
	"""
	Converts an integer or shorthand month into the full month name.

	:param month: The month number or shorthand name

	:return: The full name of the month
	"""

	error_text = f"The given month ({month!r}) is not recognised."

	try:
		month = int(month)
	except ValueError:
		try:
			return months[month.capitalize()[:3]]  # type: ignore
		except KeyError:
			raise ValueError(error_text)

	# Only get here if first try succeeded
	if 0 < month <= 12:
		return list(months.values())[month - 1]
	else:
		raise ValueError(error_text)


def get_month_number(month: Union[str, int]) -> int:
	"""
	Returns the number of the given month.
	If ``month`` is already a number between 1 and 12 it will be returned immediately.

	:param month: The month to convert to a number

	:return: The number of the month
	"""

	if isinstance(month, int):
		if 0 < month <= 12:
			return month
		else:
			raise ValueError(f"The given month ({month!r}) is not recognised.")
	else:
		month = parse_month(month)
		return list(months.values()).index(month) + 1


def check_date(month: Union[str, int], day: int, leap_year: bool = True) -> bool:
	"""
	Returns :py:obj:`True` if the day number is valid for the given month.

	.. note::

		This function will return :py:obj:`True` for the 29th Feb.
		If you don't want this behaviour set ``leap_year`` to :py:obj:`False`.

	.. latex:vspace:: -10px

	:param month: The month to test.
	:param day: The day number to test.
	:param leap_year: Whether to return :py:obj:`True` for 29th Feb.
	"""

	# Ensure day is an integer
	day = int(day)
	month = get_month_number(month)
	year = 2020 if leap_year else 2019

	try:
		datetime.date(year, month, day)
		return True
	except ValueError:
		return False


def calc_easter(year: int) -> datetime.date:
	"""
	Returns the date of Easter in the given year.

	.. versionadded:: 1.4.0

	:param year:
	"""

	a = year % 19
	b = year // 100
	c = year % 100
	d = (19 * a + b - b // 4 - ((b - (b + 8) // 25 + 1) // 3) + 15) % 30
	e = (32 + 2 * (b % 4) + 2 * (c // 4) - d - (c % 4)) % 7
	f = d + e - 7 * ((a + 11 * d + 22 * e) // 451) + 114
	month = f // 31
	day = f % 31 + 1

	return datetime.date(year, month, day)


def get_utc_offset(
		tz: Union[datetime.tzinfo, str],
		date: Optional[datetime.datetime] = None,
		) -> Optional[datetime.timedelta]:
	"""
	Returns the offset between UTC and the requested timezone on the given date.
	If ``date`` is :py:obj:`None` then the current date is used.

	:param tz: ``pytz.timezone`` or a string representing the timezone
	:param date: The date to obtain the UTC offset for
	"""

	if date is None:
		date = datetime.datetime.now(pytz.utc)

	timezone: Optional[datetime.tzinfo]

	if isinstance(tz, str):
		timezone = get_timezone(tz, date)
	else:
		timezone = tz  # pragma: no cover (hard to test)

	return date.replace(tzinfo=pytz.utc).astimezone(timezone).utcoffset()


def get_timezone(tz: str, date: Optional[datetime.datetime] = None) -> Optional[datetime.tzinfo]:
	"""
	Returns a localized ``pytz.timezone`` object for the given date.

	If ``date`` is :py:obj:`None` then the current date is used.

	.. latex:vspace:: -10px

	:param tz: A string representing a pytz timezone
	:param date: The date to obtain the timezone for
	"""

	if date is None:  # pragma: no cover (hard to test)
		date = datetime.datetime.now(pytz.utc)

	d = date.replace(tzinfo=None)

	return pytz.timezone(tz).localize(d).tzinfo


def is_bst(the_date: Union[time.struct_time, datetime.date]) -> bool:
	"""
	Calculates whether the given day falls within British Summer Time.

	This function should also be applicable to other timezones
	which change to summer time on the same date (e.g. Central European Summer Time).

	.. note::

		This function does not consider the time of day,
		and therefore does not handle the fact that the time changes at 1 AM GMT.
		It also does not account for historic deviations from the current norm.

	.. versionadded:: 3.5.0

	:param the_date: A :class:`time.struct_time`, :class:`datetime.date`
		or :class:`datetime.datetime` representing the target date.

	:returns: :py:obj:`True` if the date falls within British Summer Time, :py:obj:`False` otherwise.
	"""

	if isinstance(the_date, datetime.date):
		the_date = the_date.timetuple()

	day, month, dow = the_date.tm_mday, the_date.tm_mon, (the_date.tm_wday + 1) % 7

	if 3 > month > 10:
		return False
	elif 3 < month < 10:
		return True
	elif month == 3:
		return day - dow >= 25
	elif month == 10:
		return day - dow < 25
	else:
		return False


_pytz_functions = ["get_utc_offset", "get_timezone"]

try:

	# 3rd party
	import pytz

	__all__.extend(_pytz_functions)

except ImportError as e:

	if __name__ == "__main__":

		# stdlib
		import warnings

		# this package
		from domdf_python_tools.words import word_join

		warnings.warn(
				f"""\
		'{word_join(_pytz_functions)}' require pytz (https://pypi.org/project/pytz/), but it could not be imported.

		The error was: {e}.
		"""
				)

	else:
		_actual_module = sys.modules[__name__]

		class SelfWrapper(ModuleType):

			def __getattr__(self, name):
				if name in _pytz_functions:
					raise ImportError(
							f"{name!r} requires pytz (https://pypi.org/project/pytz/), but it could not be imported."
							)
				else:
					return getattr(_actual_module, name)

		sys.modules[__name__] = SelfWrapper(__name__)



# noqa: D100,DALL000

# stdlib
import os
import sys
from typing import Any, BinaryIO, TextIO

if sys.version_info[:2] < (3, 9):  # pragma: no cover (py39+)
	# 3rd party
	import importlib_resources

	globals().update(importlib_resources.__dict__)

else:  # pragma: no cover (<py39)
	# stdlib
	import importlib.resources
	globals().update(importlib.resources.__dict__)

if not ((3, 9) <= sys.version_info < (3, 11)):  # pragma: no cover (py39 OR py310):

	def _normalize_path(path: Any) -> str:
		"""
		Normalize a path by ensuring it is a string.

		If the resulting string contains path separators, an exception is raised.
		"""

		parent, file_name = os.path.split(str(path))
		if parent:
			raise ValueError(f'{path!r} must be only a file name')
		return file_name

	def open_binary(package: "Package", resource: "Resource") -> BinaryIO:
		"""
		Return a file-like object opened for binary reading of the resource.
		"""

		return (files(package) / _normalize_path(resource)).open("rb")

	def read_binary(package: "Package", resource: "Resource") -> bytes:
		"""
		Return the binary contents of the resource.
		"""

		return (files(package) / _normalize_path(resource)).read_bytes()

	def open_text(
			package: "Package",
			resource: "Resource",
			encoding: str = "utf-8",
			errors: str = "strict",
			) -> TextIO:
		"""
		Return a file-like object opened for text reading of the resource.
		"""

		return (files(package) / _normalize_path(resource)).open(
				'r',
				encoding=encoding,
				errors=errors,
				)

	def read_text(
			package: "Package",
			resource: "Resource",
			encoding: str = "utf-8",
			errors: str = "strict",
			) -> str:
		"""
		Return the decoded string of the resource.
		"""

		with open_text(package, resource, encoding, errors) as fp:
			return fp.read()



#!/usr/bin/env python
#
#  versions.py
"""
NamedTuple-like class to represent a version number.

.. versionadded:: 0.4.4
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import re
from typing import Dict, Generator, Iterable, Sequence, Tuple, Type, TypeVar, Union

# 3rd party
from typing_extensions import final

__all__ = ["Version"]

_V = TypeVar("_V", bound="Version")


@final
class Version(Tuple[int, int, int]):
	"""
	NamedTuple-like class to represent a version number.

	:param major:

	.. versionchanged:: 1.4.0 Implemented the same interface as a :func:`collections.namedtuple`.
	"""

	__slots__ = ()

	#: The major version number.
	major: int

	#: The minor version number.
	minor: int

	#: The patch version number.
	patch: int

	_fields: Tuple[str, str, str] = ("major", "minor", "patch")
	"""
	Tuple of strings listing the field names.

	Useful for introspection and for creating new named tuple types from existing named tuples.

	.. versionadded:: 1.4.0
	"""

	_field_defaults: Dict[str, int] = {"major": 0, "minor": 0, "patch": 0}
	"""
	Dictionary mapping field names to default values.

	.. versionadded:: 1.4.0
	"""

	@property  # type: ignore
	def major(self):  # noqa: D102
		return self[0]

	@property  # type: ignore
	def minor(self):  # noqa: D102
		return self[1]

	@property  # type: ignore
	def patch(self):  # noqa: D102
		return self[2]

	def __new__(cls: Type[_V], major=0, minor=0, patch=0) -> _V:  # noqa: D102
		t: _V = super().__new__(cls, (int(major), int(minor), int(patch)))  # type: ignore

		return t

	def __repr__(self) -> str:
		"""
		Return the representation of the version.
		"""

		repr_fmt = '(' + ", ".join(f"{name}=%r" for name in self._fields) + ')'
		return self.__class__.__name__ + repr_fmt % self

	def __str__(self) -> str:
		"""
		Return version as a string.
		"""

		return 'v' + '.'.join(str(x) for x in self)  # pylint: disable=not-an-iterable

	def __float__(self) -> float:
		"""
		Return the major and minor version number as a float.
		"""

		return float('.'.join(str(x) for x in self[:2]))

	def __int__(self) -> int:
		"""
		Return the major version number as an integer.
		"""

		return self.major

	def __getnewargs__(self):
		"""
		Return Version as a plain tuple. Used by copy and pickle.
		"""

		return tuple(self)

	def __eq__(self, other) -> bool:
		"""
		Returns whether this version is equal to the other version.

		:type other: :class:`str`, :class:`float`, :class:`~.Version`
		"""

		other = _prep_for_eq(other)

		if other is NotImplemented:
			return NotImplemented  # pragma: no cover
		else:
			shortest = min(len(self), (len(other)))
			return self[:shortest] == other[:shortest]

	def __gt__(self, other) -> bool:
		"""
		Returns whether this version is greater than the other version.

		:type other: :class:`str`, :class:`float`, :class:`~.Version`
		"""

		other = _prep_for_eq(other)

		if other is NotImplemented:
			return NotImplemented  # pragma: no cover
		else:
			return tuple(self) > other

	def __lt__(self, other) -> bool:
		"""
		Returns whether this version is less than the other version.

		:type other: :class:`str`, :class:`float`, :class:`~.Version`
		"""

		other = _prep_for_eq(other)

		if other is NotImplemented:
			return NotImplemented  # pragma: no cover
		else:
			return tuple(self) < other

	def __ge__(self, other) -> bool:
		"""
		Returns whether this version is greater than or equal to the other version.

		:type other: :class:`str`, :class:`float`, :class:`~.Version`
		"""

		other = _prep_for_eq(other)

		if other is NotImplemented:
			return NotImplemented  # pragma: no cover
		else:
			return tuple(self)[:len(other)] >= other

	def __le__(self, other) -> bool:
		"""
		Returns whether this version is less than or equal to the other version.

		:type other: :class:`str`, :class:`float`, :class:`~.Version`
		"""

		other = _prep_for_eq(other)

		if other is NotImplemented:
			return NotImplemented  # pragma: no cover
		else:
			return tuple(self)[:len(other)] <= other

	@classmethod
	def from_str(cls: Type[_V], version_string: str) -> _V:
		"""
		Create a :class:`~.Version` from a :class:`str`.

		:param version_string: The version number.

		:return: The created :class:`~domdf_python_tools.versions.Version`.
		"""

		return cls(*_iter_string(version_string))

	@classmethod
	def from_tuple(cls: Type[_V], version_tuple: Tuple[Union[str, int], ...]) -> _V:
		"""
		Create a :class:`~.Version` from a :class:`tuple`.

		:param version_tuple: The version number.

		:return: The created :class:`~domdf_python_tools.versions.Version`.

		.. versionchanged:: 0.9.0

			Tuples with more than three elements are truncated.
			Previously a :exc:`TypeError` was raised.
		"""

		return cls(*(int(x) for x in version_tuple[:3]))

	@classmethod
	def from_float(cls: Type[_V], version_float: float) -> _V:
		"""
		Create a :class:`~.Version` from a :class:`float`.

		:param version_float: The version number.

		:return: The created :class:`~domdf_python_tools.versions.Version`.
		"""

		return cls.from_str(str(version_float))

	def _asdict(self) -> Dict[str, int]:
		"""
		Return a new dict which maps field names to their corresponding values.

		.. versionadded:: 1.4.0
		"""

		return {
				"major": self.major,
				"minor": self.minor,
				"patch": self.patch,
				}

	def _replace(self: _V, **kwargs) -> _V:
		"""
		Return a new instance of the named tuple replacing specified fields with new values.

		.. versionadded:: 1.4.0

		:param kwargs:
		"""

		return self.__class__(**{**self._asdict(), **kwargs})

	@classmethod
	def _make(cls: Type[_V], iterable: Iterable[Union[str, int]]) -> _V:
		"""
		Class method that makes a new instance from an existing sequence or iterable.

		.. versionadded:: 1.4.0

		:param iterable:
		"""

		return cls(*(int(x) for x in tuple(iterable)[:3]))


def _iter_string(version_string: str) -> Generator[int, None, None]:
	"""
	Iterate over the version elements from a string.

	:param version_string: The version as a string.

	:return: Iterable elements of the version.
	"""

	return (int(x) for x in re.split("[.,]", version_string))


def _iter_float(version_float: float) -> Generator[int, None, None]:
	"""
	Iterate over the version elements from a float.

	:param version_float: The version as a float.

	:return: Iterable elements of the version.
	"""

	return _iter_string(str(version_float))


def _prep_for_eq(other: Union[str, float, Version], ) -> Tuple[int, ...]:
	"""
	Prepare 'other' for use in ``__eq__``, ``__le__``, ``__ge__``, ``__gt__``, and ``__lt__``.
	"""

	if isinstance(other, str):
		return tuple(_iter_string(other))
	elif isinstance(other, (Version, Sequence)):
		return tuple(int(x) for x in other)
	elif isinstance(other, (int, float)):
		return tuple(_iter_float(other))
	else:  # pragma: no cover
		return NotImplemented



# From https://github.com/python/typeshed
# Apache-2.0 Licensed

# stdlib
import abc
import pathlib
import sys
from collections.abc import Mapping
from email.message import Message
from importlib.abc import MetaPathFinder
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple, Union, overload

StrPath = Union[str, PathLike[str]]

if sys.version_info >= (3, 10):

	def packages_distributions() -> Mapping[str, List[str]]: ...

class PackageNotFoundError(ModuleNotFoundError): ...

class _EntryPointBase(NamedTuple):
	name: str
	value: str
	group: str

class EntryPoint(_EntryPointBase):
	def load(self) -> Any: ...  # Callable[[], Any] or an importable module

	@property
	def extras(self) -> List[str]: ...

class PackagePath(pathlib.PurePosixPath):
	def read_text(self, encoding: str = ...) -> str: ...
	def read_binary(self) -> bytes: ...
	def locate(self) -> PathLike[str]: ...

	# The following attributes are not defined on PackagePath, but are dynamically added by Distribution.files:
	hash: Optional[FileHash]  # noqa: A003  # pylint: disable=redefined-builtin
	size: Optional[int]
	dist: Distribution

class FileHash:
	mode: str
	value: str

	def __init__(self, spec: str) -> None: ...

class Distribution:

	@abc.abstractmethod
	def read_text(self, filename: str) -> Optional[str]: ...

	@abc.abstractmethod
	def locate_file(self, path: StrPath) -> PathLike[str]: ...

	@classmethod
	def from_name(cls, name: str) -> Distribution: ...

	@overload
	@classmethod
	def discover(cls, *, context: DistributionFinder.Context) -> Iterable[Distribution]: ...

	@overload
	@classmethod
	def discover(
			cls,
			*,
			context: None = ...,
			name: Optional[str] = ...,
			path: List[str] = ...,
			**kwargs: Any,
			) -> Iterable[Distribution]: ...

	@staticmethod
	def at(path: StrPath) -> PathDistribution: ...

	@property
	def metadata(self) -> Message: ...

	@property
	def version(self) -> str: ...

	@property
	def entry_points(self) -> List[EntryPoint]: ...

	@property
	def files(self) -> Optional[List[PackagePath]]: ...

	@property
	def requires(self) -> Optional[List[str]]: ...

class DistributionFinder(MetaPathFinder):

	class Context:
		name: Optional[str]

		def __init__(self, *, name: Optional[str] = ..., path: List[str] = ..., **kwargs: Any) -> None: ...

		@property
		def path(self) -> List[str]: ...

	@abc.abstractmethod
	def find_distributions(self, context: DistributionFinder.Context = ...) -> Iterable[Distribution]: ...

class MetadataPathFinder(DistributionFinder):

	@classmethod
	def find_distributions(cls, context: DistributionFinder.Context = ...) -> Iterable[PathDistribution]: ...

class PathDistribution(Distribution):
	def __init__(self, path: Path) -> None: ...
	def read_text(self, filename: StrPath) -> str: ...
	def locate_file(self, path: StrPath) -> PathLike[str]: ...

def distribution(distribution_name: str) -> Distribution: ...

@overload
def distributions(*, context: DistributionFinder.Context) -> Iterable[Distribution]: ...

@overload
def distributions(
		*,
		context: None = ...,
		name: Optional[str] = ...,
		path: List[str] = ...,
		**kwargs: Any,
		) -> Iterable[Distribution]: ...

def metadata(distribution_name: str) -> Message: ...
def version(distribution_name: str) -> str: ...
def entry_points() -> Dict[str, Tuple[EntryPoint, ...]]: ...
def files(distribution_name: str) -> Optional[List[PackagePath]]: ...
def requires(distribution_name: str) -> Optional[List[str]]: ...



#!/usr/bin/env python
#
#  paths.py
"""
Functions for paths and files.

.. versionchanged:: 1.0.0

	Removed ``relpath2``.
	Use :func:`domdf_python_tools.paths.relpath` instead.
"""
#
#  Copyright © 2018-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Parts of the docstrings, the PathPlus class and the DirComparator class
#  based on Python and its Documentation
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2021 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#
#  copytree based on https://stackoverflow.com/a/12514470/3092681
#      Copyright © 2012 atzz
#      Licensed under CC-BY-SA
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import contextlib
import filecmp
import fnmatch
import gzip
import json
import os
import pathlib
import shutil
import stat
import sys
import tempfile
import urllib.parse
from collections import defaultdict, deque
from operator import methodcaller
from typing import (
		IO,
		Any,
		Callable,
		ContextManager,
		Dict,
		Iterable,
		Iterator,
		List,
		Optional,
		Sequence,
		Type,
		TypeVar,
		Union
		)

# this package
from domdf_python_tools.compat import nullcontext
from domdf_python_tools.typing import JsonLibrary, PathLike

__all__ = [
		"append",
		"copytree",
		"delete",
		"maybe_make",
		"parent_path",
		"read",
		"relpath",
		"write",
		"clean_writer",
		"make_executable",
		"PathPlus",
		"PosixPathPlus",
		"WindowsPathPlus",
		"in_directory",
		"_P",
		"_PP",
		"traverse_to_file",
		"matchglob",
		"unwanted_dirs",
		"TemporaryPathPlus",
		"sort_paths",
		"DirComparator",
		"compare_dirs",
		]

NEWLINE_DEFAULT = type("NEWLINE_DEFAULT", (object, ), {"__repr__": lambda self: "NEWLINE_DEFAULT"})()

_P = TypeVar("_P", bound=pathlib.Path)
"""
.. versionadded:: 0.11.0

.. versionchanged:: 1.7.0  Now bound to :class:`pathlib.Path`.
"""

_PP = TypeVar("_PP", bound="PathPlus")
"""
.. versionadded:: 2.3.0
"""

unwanted_dirs = (
		".git",
		".hg",
		"venv",
		".venv",
		".mypy_cache",
		"__pycache__",
		".pytest_cache",
		".tox",
		".tox4",
		".nox",
		"__pypackages__",
		"dosdevices",
		)
"""
A list of directories which will likely be unwanted when searching directory trees for files.

.. versionadded:: 2.3.0
.. versionchanged:: 2.9.0  Added ``.hg`` (`mercurial <https://www.mercurial-scm.org>`_)
.. versionchanged:: 3.0.0  Added ``__pypackages__`` (:pep:`582`)
.. versionchanged:: 3.2.0  Added ``.nox`` (https://nox.thea.codes/)
"""


def append(var: str, filename: PathLike, **kwargs) -> int:
	"""
	Append ``var`` to the file ``filename`` in the current directory.

	.. TODO:: make this the file in the given directory, by default the current directory

	:param var: The value to append to the file
	:param filename: The file to append to
	"""

	kwargs.setdefault("encoding", "UTF-8")

	with open(os.path.join(os.getcwd(), filename), 'a', **kwargs) as f:  # noqa: ENC001
		return f.write(var)


def copytree(
		src: PathLike,
		dst: PathLike,
		symlinks: bool = False,
		ignore: Optional[Callable] = None,
		) -> PathLike:
	"""
	Alternative to :func:`shutil.copytree` to support copying to a directory that already exists.

	Based on https://stackoverflow.com/a/12514470 by https://stackoverflow.com/users/23252/atzz

	In Python 3.8 and above :func:`shutil.copytree` takes a ``dirs_exist_ok`` argument,
	which has the same result.

	:param src: Source file to copy
	:param dst: Destination to copy file to
	:param symlinks: Whether to represent symbolic links in the source as symbolic
		links in the destination. If false or omitted, the contents and metadata
		of the linked files are copied to the new tree. When symlinks is false,
		if the file pointed by the symlink doesn't exist, an exception will be
		added in the list of errors raised in an Error exception at the end of
		the copy process. You can set the optional ignore_dangling_symlinks
		flag to true if you want to silence this exception. Notice that this
		option has no effect on platforms that don’t support :func:`os.symlink`.
	:param ignore: A callable that will receive as its arguments the source
		directory, and a list of its contents. The ignore callable will be
		called once for each directory that is copied. The callable must return
		a sequence of directory and file names relative to the current
		directory (i.e. a subset of the items in its second argument); these
		names will then be ignored in the copy process.
		:func:`shutil.ignore_patterns` can be used to create such a callable
		that ignores names based on
		glob-style patterns.
	"""

	for item in os.listdir(src):
		s = os.path.join(src, item)
		d = os.path.join(dst, item)
		if os.path.isdir(s):
			shutil.copytree(s, d, symlinks, ignore)
		else:
			shutil.copy2(s, d)

	return dst


def delete(filename: PathLike, **kwargs):
	"""
	Delete the file in the current directory.

	.. TODO:: make this the file in the given directory, by default the current directory

	:param filename: The file to delete
	"""

	os.remove(os.path.join(os.getcwd(), filename), **kwargs)


def maybe_make(directory: PathLike, mode: int = 0o777, parents: bool = False):
	"""
	Create a directory at the given path, but only if the directory does not already exist.

	.. attention::

		This will fail silently if a file with the same name already exists.
		This appears to be due to the behaviour of :func:`os.mkdir`.

	:param directory: Directory to create
	:param mode: Combined with the process's umask value to determine the file mode and access flags
	:param parents: If :py:obj:`False` (the default), a missing parent raises a :class:`FileNotFoundError`.
		If :py:obj:`True`, any missing parents of this path are created as needed; they are created with the
		default permissions without taking mode into account (mimicking the POSIX ``mkdir -p`` command).
	:no-default parents:

	.. versionchanged:: 1.6.0  Removed the ``'exist_ok'`` option, since it made no sense in this context.

	"""

	if not isinstance(directory, pathlib.Path):
		directory = pathlib.Path(directory)

	try:
		directory.mkdir(mode, parents, exist_ok=True)
	except FileExistsError:
		pass


def parent_path(path: PathLike) -> pathlib.Path:
	"""
	Returns the path of the parent directory for the given file or directory.

	:param path: Path to find the parent for

	:return: The parent directory
	"""

	if not isinstance(path, pathlib.Path):
		path = pathlib.Path(path)

	return path.parent


def read(filename: PathLike, **kwargs) -> str:
	"""
	Read a file in the current directory (in text mode).

	.. TODO:: make this the file in the given directory, by default the current directory

	:param filename: The file to read from.

	:return: The contents of the file.
	"""

	kwargs.setdefault("encoding", "UTF-8")

	with open(os.path.join(os.getcwd(), filename), **kwargs) as f:  # noqa: ENC001
		return f.read()


def relpath(path: PathLike, relative_to: Optional[PathLike] = None) -> pathlib.Path:
	"""
	Returns the path for the given file or directory relative to the given
	directory or, if that would require path traversal, returns the absolute path.

	:param path: Path to find the relative path for
	:param relative_to: The directory to find the path relative to.
		Defaults to the current directory.
	:no-default relative_to:
	"""  # noqa: D400

	if not isinstance(path, pathlib.Path):
		path = pathlib.Path(path)

	abs_path = path.absolute()

	if relative_to is None:
		relative_to = pathlib.Path().absolute()

	if not isinstance(relative_to, pathlib.Path):
		relative_to = pathlib.Path(relative_to)

	relative_to = relative_to.absolute()

	try:
		return abs_path.relative_to(relative_to)
	except ValueError:
		return abs_path


def write(var: str, filename: PathLike, **kwargs) -> None:
	"""
	Write a variable to file in the current directory.

	.. TODO:: make this the file in the given directory, by default the current directory

	:param var: The value to write to the file.
	:param filename: The file to write to.
	"""

	kwargs.setdefault("encoding", "UTF-8")

	with open(os.path.join(os.getcwd(), filename), 'w', **kwargs) as f:  # noqa: ENC001
		f.write(var)


def clean_writer(string: str, fp: IO) -> None:
	"""
	Write string to ``fp`` without trailing spaces.

	:param string:
	:param fp:
	"""

	# this package
	from domdf_python_tools.stringlist import StringList

	buffer = StringList(string)
	buffer.blankline(ensure_single=True)
	fp.write(str(buffer))


def make_executable(filename: PathLike) -> None:
	"""
	Make the given file executable.

	:param filename:
	"""

	if not isinstance(filename, pathlib.Path):
		filename = pathlib.Path(filename)

	st = os.stat(str(filename))
	os.chmod(str(filename), st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


@contextlib.contextmanager
def in_directory(directory: PathLike):
	"""
	Context manager to change into the given directory for the
	duration of the ``with`` block.

	:param directory:
	"""  # noqa: D400

	oldwd = os.getcwd()
	try:
		os.chdir(str(directory))
		yield
	finally:
		os.chdir(oldwd)


class PathPlus(pathlib.Path):
	"""
	Subclass of :class:`pathlib.Path` with additional methods and a default encoding of UTF-8.

	Path represents a filesystem path but, unlike :class:`pathlib.PurePath`, also offers
	methods to do system calls on path objects.
	Depending on your system, instantiating a :class:`~.PathPlus` will return
	either a :class:`~.PosixPathPlus` or a :class:`~.WindowsPathPlus`. object.
	You can also instantiate a :class:`~.PosixPathPlus` or :class:`WindowsPath` directly,
	but cannot instantiate a :class:`~.WindowsPathPlus` on a POSIX system or vice versa.

	.. versionadded:: 0.3.8
	.. versionchanged:: 0.5.1  Defaults to Unix line endings (``LF``) on all platforms.
	"""

	__slots__ = ()

	if sys.version_info < (3, 11):
		_accessor = pathlib._normal_accessor  # type: ignore
	_closed = False

	def _init(self, *args, **kwargs):
		pass

	@classmethod
	def _from_parts(cls, args, init=True):
		return super()._from_parts(args)  # type: ignore

	def __new__(cls: Type[_PP], *args, **kwargs) -> _PP:  # noqa: D102
		if cls is PathPlus:
			cls = WindowsPathPlus if os.name == "nt" else PosixPathPlus  # type: ignore

		return super().__new__(cls, *args, **kwargs)

	def make_executable(self) -> None:
		"""
		Make the file executable.

		.. versionadded:: 0.3.8
		"""

		make_executable(self)

	def write_clean(
			self,
			string: str,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			):
		"""
		Write to the file without trailing whitespace, and with a newline at the end of the file.

		.. versionadded:: 0.3.8

		:param string:
		:param encoding: The encoding to write to the file in.
		:param errors:
		"""

		with self.open('w', encoding=encoding, errors=errors) as fp:
			clean_writer(string, fp)

	def maybe_make(
			self,
			mode: int = 0o777,
			parents: bool = False,
			):
		"""
		Create a directory at this path, but only if the directory does not already exist.

		.. versionadded:: 0.3.8

		:param mode: Combined with the process’ umask value to determine the file mode and access flags
		:param parents: If :py:obj:`False` (the default), a missing parent raises a :class:`FileNotFoundError`.
			If :py:obj:`True`, any missing parents of this path are created as needed; they are created with the
			default permissions without taking mode into account (mimicking the POSIX mkdir -p command).
		:no-default parents:

		.. versionchanged:: 1.6.0  Removed the ``'exist_ok'`` option, since it made no sense in this context.

		.. attention::

			This will fail silently if a file with the same name already exists.
			This appears to be due to the behaviour of :func:`os.mkdir`.

		"""

		try:
			self.mkdir(mode, parents, exist_ok=True)
		except FileExistsError:
			pass

	def append_text(
			self,
			string: str,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			):
		"""
		Open the file in text mode, append the given string to it, and close the file.

		.. versionadded:: 0.3.8

		:param string:
		:param encoding: The encoding to write to the file in.
		:param errors:
		"""

		with self.open('a', encoding=encoding, errors=errors) as fp:
			fp.write(string)

	def write_text(
			self,
			data: str,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			newline: Optional[str] = NEWLINE_DEFAULT,
			) -> int:
		"""
		Open the file in text mode, write to it, and close the file.

		.. versionadded:: 0.3.8

		:param data:
		:param encoding: The encoding to write to the file in.
		:param errors:
		:param newline:
		:default newline: `universal newlines <https://docs.python.org/3/glossary.html#term-universal-newlines>`__ for reading, Unix line endings (``LF``) for writing.

		.. versionchanged:: 3.1.0

			Added the ``newline`` argument to match Python 3.10.
			(see :github:pull:`22420 <python/cpython>`)
		"""

		if not isinstance(data, str):
			raise TypeError(f'data must be str, not {data.__class__.__name__}')

		with self.open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
			return f.write(data)

	def write_lines(
			self,
			data: Iterable[str],
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			*,
			trailing_whitespace: bool = False
			) -> None:
		"""
		Write the given list of lines to the file without trailing whitespace.

		.. versionadded:: 0.5.0

		:param data:
		:param encoding: The encoding to write to the file in.
		:param errors:
		:param trailing_whitespace: If :py:obj:`True` trailing whitespace is preserved.

		.. versionchanged:: 2.4.0  Added the ``trailing_whitespace`` option.
		"""

		if trailing_whitespace:
			data = list(data)
			if data[-1].strip():
				data.append('')

			self.write_text('\n'.join(data), encoding=encoding, errors=errors)
		else:
			self.write_clean('\n'.join(data), encoding=encoding, errors=errors)

	def read_text(
			self,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			) -> str:
		"""
		Open the file in text mode, read it, and close the file.

		.. versionadded:: 0.3.8

		:param encoding: The encoding to write to the file in.
		:param errors:

		:return: The content of the file.
		"""

		return super().read_text(encoding=encoding, errors=errors)

	def read_lines(
			self,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			) -> List[str]:
		"""
		Open the file in text mode, return a list containing the lines in the file,
		and close the file.

		.. versionadded:: 0.5.0

		:param encoding: The encoding to write to the file in.
		:param errors:

		:return: The content of the file.
		"""  # noqa: D400

		return self.read_text(encoding=encoding, errors=errors).split('\n')

	def open(  # type: ignore  # noqa: A003  # pylint: disable=redefined-builtin
		self,
		mode: str = 'r',
		buffering: int = -1,
		encoding: Optional[str] = "UTF-8",
		errors: Optional[str] = None,
		newline: Optional[str] = NEWLINE_DEFAULT,
		) -> IO[Any]:
		"""
		Open the file pointed by this path and return a file object, as
		the built-in :func:`open` function does.

		.. versionadded:: 0.3.8

		:param mode: The mode to open the file in.
		:default mode: ``'r'`` (read only)
		:param buffering:
		:param encoding:
		:param errors:
		:param newline:
		:default newline: `universal newlines <https://docs.python.org/3/glossary.html#term-universal-newlines>`__ for reading, Unix line endings (``LF``) for writing.

		:rtype:

		.. versionchanged:: 0.5.1

			Defaults to Unix line endings (``LF``) on all platforms.
		"""  # noqa: D400

		if 'b' in mode:
			encoding = None
			newline = None

		if newline is NEWLINE_DEFAULT:
			if 'r' in mode:
				newline = None
			else:
				newline = '\n'

		return super().open(
				mode,
				buffering=buffering,
				encoding=encoding,
				errors=errors,
				newline=newline,
				)

	def dump_json(
			self,
			data: Any,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			json_library: JsonLibrary = json,  # type: ignore
			*,
			compress: bool = False,
			**kwargs,
			) -> None:
		r"""
		Dump ``data`` to the file as JSON.

		.. versionadded:: 0.5.0

		:param data: The object to serialise to JSON.
		:param encoding: The encoding to write to the file in.
		:param errors:
		:param json_library: The JSON serialisation library to use.
		:default json_library: :mod:`json`
		:param compress: Whether to compress the JSON file using gzip.
		:param \*\*kwargs: Keyword arguments to pass to the JSON serialisation function.

		:rtype:

		.. versionchanged:: 1.0.0

			Now uses :meth:`PathPlus.write_clean <domdf_python_tools.paths.PathPlus.write_clean>`
			rather than :meth:`PathPlus.write_text <domdf_python_tools.paths.PathPlus.write_text>`,
			and as a result returns :py:obj:`None` rather than :class:`int`.

		.. versionchanged:: 1.9.0  Added the ``compress`` keyword-only argument.
		"""

		if compress:
			with gzip.open(self, mode="wt", encoding=encoding, errors=errors) as fp:
				fp.write(json_library.dumps(data, **kwargs))

		else:
			self.write_clean(
					json_library.dumps(data, **kwargs),
					encoding=encoding,
					errors=errors,
					)

	def load_json(
			self,
			encoding: Optional[str] = "UTF-8",
			errors: Optional[str] = None,
			json_library: JsonLibrary = json,  # type: ignore
			*,
			decompress: bool = False,
			**kwargs,
			) -> Any:
		r"""
		Load JSON data from the file.

		.. versionadded:: 0.5.0

		:param encoding: The encoding to write to the file in.
		:param errors:
		:param json_library: The JSON serialisation library to use.
		:default json_library: :mod:`json`
		:param decompress: Whether to decompress the JSON file using gzip.
			Will raise an exception if the file is not compressed.
		:param \*\*kwargs: Keyword arguments to pass to the JSON deserialisation function.

		:return: The deserialised JSON data.

		.. versionchanged:: 1.9.0  Added the ``compress`` keyword-only argument.
		"""

		if decompress:
			with gzip.open(self, mode="rt", encoding=encoding, errors=errors) as fp:
				content = fp.read()
		else:
			content = self.read_text(encoding=encoding, errors=errors)

		return json_library.loads(
				content,
				**kwargs,
				)

	if sys.version_info < (3, 10):  # pragma: no cover (py310+)

		def is_mount(self) -> bool:
			"""
			Check if this path is a POSIX mount point.

			.. versionadded:: 0.3.8 for Python 3.7 and above
			.. versionadded:: 0.11.0 for Python 3.6
			"""

			# Need to exist and be a dir
			if not self.exists() or not self.is_dir():
				return False

			# https://github.com/python/cpython/pull/18839/files
			try:
				parent_dev = self.parent.stat().st_dev
			except OSError:
				return False

			dev = self.stat().st_dev
			if dev != parent_dev:
				return True
			ino = self.stat().st_ino
			parent_ino = self.parent.stat().st_ino
			return ino == parent_ino

	if sys.version_info < (3, 8):  # pragma: no cover (py38+)

		def rename(self: _P, target: Union[str, pathlib.PurePath]) -> _P:
			"""
			Rename this path to the target path.

			The target path may be absolute or relative. Relative paths are
			interpreted relative to the current working directory, *not* the
			directory of the Path object.

			.. versionadded:: 0.3.8 for Python 3.8 and above
			.. versionadded:: 0.11.0 for Python 3.6 and Python 3.7

			:param target:

			:returns: The new Path instance pointing to the target path.
			"""

			os.rename(self, target)
			return self.__class__(target)

		def replace(self: _P, target: Union[str, pathlib.PurePath]) -> _P:
			"""
			Rename this path to the target path, overwriting if that path exists.

			The target path may be absolute or relative. Relative paths are
			interpreted relative to the current working directory, *not* the
			directory of the Path object.

			Returns the new Path instance pointing to the target path.

			.. versionadded:: 0.3.8 for Python 3.8 and above
			.. versionadded:: 0.11.0 for Python 3.6 and Python 3.7

			:param target:

			:returns: The new Path instance pointing to the target path.
			"""

			os.replace(self, target)
			return self.__class__(target)

		def unlink(self, missing_ok: bool = False) -> None:
			"""
			Remove this file or link.

			If the path is a directory, use :meth:`~domdf_python_tools.paths.PathPlus.rmdir()` instead.

			.. versionadded:: 0.3.8 for Python 3.8 and above
			.. versionadded:: 0.11.0 for Python 3.6 and Python 3.7
			"""

			try:
				os.unlink(self)
			except FileNotFoundError:
				if not missing_ok:
					raise

	def __enter__(self):
		return self

	def __exit__(self, t, v, tb):
		# https://bugs.python.org/issue39682
		# In previous versions of pathlib, this method marked this path as
		# closed; subsequent attempts to perform I/O would raise an IOError.
		# This functionality was never documented, and had the effect of
		# making Path objects mutable, contrary to PEP 428. In Python 3.9 the
		# _closed attribute was removed, and this method made a no-op.
		# This method and __enter__()/__exit__() should be deprecated and
		# removed in the future.
		pass

	if sys.version_info < (3, 9):  # pragma: no cover (py39+)

		def is_relative_to(self, *other: Union[str, os.PathLike]) -> bool:
			r"""
			Returns whether the path is relative to another path.

			.. versionadded:: 0.3.8 for Python 3.9 and above.
			.. latex:vspace:: -10px
			.. versionadded:: 1.4.0 for Python 3.6 and Python 3.7.
			.. latex:vspace:: -10px

			:param \*other:

			.. latex:vspace:: -20px

			:rtype:

			.. latex:vspace:: -20px
			"""

			try:
				self.relative_to(*other)
				return True
			except ValueError:
				return False

	def abspath(self) -> "PathPlus":
		"""
		Return the absolute version of the path.

		.. versionadded:: 1.3.0
		"""

		return self.__class__(os.path.abspath(self))

	def iterchildren(
			self: _PP,
			exclude_dirs: Optional[Iterable[str]] = unwanted_dirs,
			match: Optional[str] = None,
			matchcase: bool = True,
			) -> Iterator[_PP]:
		"""
		Returns an iterator over all children (files and directories) of the current path object.

		.. versionadded:: 2.3.0

		:param exclude_dirs: A list of directory names which should be excluded from the output,
			together with their children.
		:param match: A pattern to match filenames against.
			The pattern should be in the format taken by :func:`~.matchglob`.
		:param matchcase: Whether the filename's case should match the pattern.

		:rtype:

		.. versionchanged:: 2.5.0  Added the ``matchcase`` option.
		"""

		if not self.abspath().is_dir():
			return

		if exclude_dirs is None:
			exclude_dirs = ()

		if match and not os.path.isabs(match) and self.is_absolute():
			match = (self / match).as_posix()

		file: _PP
		for file in self.iterdir():
			parts = file.parts
			if any(d in parts for d in exclude_dirs):
				continue

			if match is None or (match is not None and matchglob(file, match, matchcase)):
				yield file

			if file.is_dir():
				yield from file.iterchildren(exclude_dirs, match)

	@classmethod
	def from_uri(cls: Type[_PP], uri: str) -> _PP:
		"""
		Construct a :class:`~.PathPlus` from a ``file`` URI returned by :meth:`pathlib.PurePath.as_uri`.

		.. versionadded:: 2.9.0

		:param uri:

		:rtype: :class:`~.PathPlus`
		"""

		parseresult = urllib.parse.urlparse(uri)

		if parseresult.scheme != "file":
			raise ValueError(f"Unsupported URI scheme {parseresult.scheme!r}")
		if parseresult.params or parseresult.query or parseresult.fragment:
			raise ValueError("Malformed file URI")

		if sys.platform == "win32":  # pragma: no cover (!Windows)

			if parseresult.netloc:
				path = ''.join([
						"//",
						urllib.parse.unquote_to_bytes(parseresult.netloc).decode("UTF-8"),
						urllib.parse.unquote_to_bytes(parseresult.path).decode("UTF-8"),
						])
			else:
				path = urllib.parse.unquote_to_bytes(parseresult.path).decode("UTF-8").lstrip('/')

		else:  # pragma: no cover (Windows)
			if parseresult.netloc:
				raise ValueError("Malformed file URI")

			path = urllib.parse.unquote_to_bytes(parseresult.path).decode("UTF-8")

		return cls(path)

	def move(self: _PP, dst: PathLike) -> _PP:
		"""
		Recursively move ``self`` to ``dst``.

		``self`` may be a file or a directory.

		See :func:`shutil.move` for more details.

		.. versionadded:: 3.2.0

		:param dst:

		:returns: The new location of ``self``.
		:rtype: :class:`~.PathPlus`
		"""

		new_path = shutil.move(os.fspath(self), dst)
		return self.__class__(new_path)

	def stream(self, chunk_size: int = 1024) -> Iterator[bytes]:
		"""
		Stream the file in ``chunk_size`` sized chunks.

		:param chunk_size: The chunk size, in bytes

		.. versionadded:: 3.2.0
		"""

		with self.open("rb") as fp:
			while True:
				chunk = fp.read(chunk_size)
				if not chunk:
					break
				yield chunk


class PosixPathPlus(PathPlus, pathlib.PurePosixPath):
	"""
	:class:`~.PathPlus` subclass for non-Windows systems.

	On a POSIX system, instantiating a :class:`~.PathPlus` object should return an instance of this class.

	.. versionadded:: 0.3.8
	"""

	__slots__ = ()


class WindowsPathPlus(PathPlus, pathlib.PureWindowsPath):
	"""
	:class:`~.PathPlus` subclass for Windows systems.

	On a Windows system, instantiating a :class:`~.PathPlus`  object should return an instance of this class.

	.. versionadded:: 0.3.8

	.. autoclasssumm:: WindowsPathPlus
		:autosummary-sections: ;;

	The following methods are unsupported on Windows:

	* :meth:`~pathlib.Path.group`
	* :meth:`~pathlib.Path.is_mount`
	* :meth:`~pathlib.Path.owner`
	"""

	__slots__ = ()

	def owner(self):  # pragma: no cover
		"""
		Unsupported on Windows.
		"""

		raise NotImplementedError("Path.owner() is unsupported on this system")

	def group(self):  # pragma: no cover
		"""
		Unsupported on Windows.
		"""

		raise NotImplementedError("Path.group() is unsupported on this system")

	def is_mount(self):  # pragma: no cover
		"""
		Unsupported on Windows.
		"""

		raise NotImplementedError("Path.is_mount() is unsupported on this system")


def traverse_to_file(base_directory: _P, *filename: PathLike, height: int = -1) -> _P:
	r"""
	Traverse the parents of the given directory until the desired file is found.

	.. versionadded:: 1.7.0

	:param base_directory: The directory to start searching from
	:param \*filename: The filename(s) to search for
	:param height: The maximum height to traverse to.
	"""

	if not filename:
		raise TypeError("traverse_to_file expected 2 or more arguments, got 1")

	for level, directory in enumerate((base_directory, *base_directory.parents)):
		if height > 0 and ((level - 1) > height):
			break

		for file in filename:
			if (directory / file).is_file():
				return directory

	raise FileNotFoundError(f"'{filename[0]!s}' not found in {base_directory}")


def matchglob(filename: PathLike, pattern: str, matchcase: bool = True) -> bool:
	"""
	Given a filename and a glob pattern, return whether the filename matches the glob.

	.. versionadded:: 2.3.0

	:param filename:
	:param pattern: A pattern structured like a filesystem path, where each element consists of the glob syntax.
		Each element is matched by :mod:`fnmatch`.
		The special element ``**`` matches zero or more files or directories.
	:param matchcase: Whether the filename's case should match the pattern.

	:rtype:

	.. seealso:: :wikipedia:`Glob (programming)#Syntax` on Wikipedia
	.. versionchanged:: 2.5.0  Added the ``matchcase`` option.
	"""

	match_func = fnmatch.fnmatchcase if matchcase else fnmatch.fnmatch

	filename = PathPlus(filename)

	pattern_parts = deque(pathlib.PurePath(pattern).parts)
	filename_parts = deque(filename.parts)

	if not pattern_parts[-1]:
		pattern_parts.pop()

	while True:
		if not pattern_parts and not filename_parts:
			return True
		elif not pattern_parts and filename_parts:
			# Pattern exhausted but still filename elements
			return False

		pattern_part = pattern_parts.popleft()

		if pattern_part == "**" and not filename_parts:
			return True
		else:
			filename_part = filename_parts.popleft()

		if pattern_part == "**":
			if not pattern_parts:
				return True

			while pattern_part == "**":
				if not pattern_parts:
					return True

				pattern_part = pattern_parts.popleft()

			if pattern_parts and not filename_parts:
				# Filename must match everything after **
				return False

			if match_func(filename_part, pattern_part):
				continue
			else:
				while not match_func(filename_part, pattern_part):
					if not filename_parts:
						return False

					filename_part = filename_parts.popleft()

		elif match_func(filename_part, pattern_part):
			continue
		else:
			return False


class TemporaryPathPlus(tempfile.TemporaryDirectory):
	"""
	Securely creates a temporary directory using the same rules as :func:`tempfile.mkdtemp`.
	The resulting object can be used as a context manager.
	On completion of the context or destruction of the object
	the newly created temporary directory and all its contents are removed from the filesystem.

	Unlike :func:`tempfile.TemporaryDirectory` this class is based around a :class:`~.PathPlus` object.

	.. versionadded:: 2.4.0
	.. autosummary-widths:: 6/16
	"""

	name: PathPlus
	"""
	The temporary directory itself.

	This will be assigned to the target of the :keyword:`as` clause if the :class:`~.TemporaryPathPlus`
	is used as a context manager.
	"""

	def __init__(
			self,
			suffix: Optional[str] = None,
			prefix: Optional[str] = None,
			dir: Optional[PathLike] = None,  # noqa: A002  # pylint: disable=redefined-builtin
			) -> None:

		super().__init__(suffix, prefix, dir)
		self.name = PathPlus(self.name)

	def cleanup(self) -> None:
		"""
		Cleanup the temporary directory by removing it and its contents.

		If the :class:`~.TemporaryPathPlus` is used as a context manager
		this is called when leaving the :keyword:`with` block.
		"""

		context: ContextManager

		if sys.platform == "win32":  # pragma: no cover (!Windows)
			context = contextlib.suppress(PermissionError, NotADirectoryError)
		else:  # pragma: no cover (Windows)
			context = nullcontext()

		with context:
			super().cleanup()

	def __enter__(self) -> PathPlus:
		return self.name


def sort_paths(*paths: PathLike) -> List[PathPlus]:
	"""
	Sort the ``paths`` by directory, then by file.

	.. versionadded:: 2.6.0

	:param paths:
	"""

	directories: Dict[str, List[PathPlus]] = defaultdict(list)
	local_contents: List[PathPlus] = []
	files: List[PathPlus] = []

	for obj in [PathPlus(path) for path in paths]:
		if len(obj.parts) > 1:
			key = obj.parts[0]
			directories[key].append(obj)
		else:
			local_contents.append(obj)

	# sort directories
	directories = {directory: directories[directory] for directory in sorted(directories.keys())}

	for directory, contents in directories.items():
		contents = [path.relative_to(directory) for path in contents]
		files.extend(PathPlus(directory) / path for path in sort_paths(*contents))

	return files + sorted(local_contents, key=methodcaller("as_posix"))


class DirComparator(filecmp.dircmp):
	r"""
	Compare the content of ``a`` and ``a``.

	In contrast with :class:`filecmp.dircmp`, this
	subclass compares the content of files with the same path.

	.. versionadded:: 2.7.0

	:param a: The "left" directory to compare.
	:param b: The "right" directory to compare.
	:param ignore: A list of names to ignore.
	:default ignore: :py:obj:`filecmp.DEFAULT_IGNORES`
	:param hide: A list of names to hide.
	:default hide: ``[`` :py:obj:`os.curdir`, :py:obj:`os.pardir` ``]``
	"""

	# From https://stackoverflow.com/a/24860799, public domain.
	# Thanks Philippe

	def __init__(
			self,
			a: PathLike,
			b: PathLike,
			ignore: Optional[Sequence[str]] = None,
			hide: Optional[Sequence[str]] = None,
			):
		super().__init__(a, b, ignore=ignore, hide=hide)

	def phase3(self) -> None:  # noqa: D102
		# Find out differences between common files.
		# Ensure we are using content comparison with shallow=False.

		fcomp = filecmp.cmpfiles(self.left, self.right, self.common_files, shallow=False)
		self.same_files, self.diff_files, self.funny_files = fcomp

	def phase4(self) -> None:  # noqa: D102
		# Find out differences between common subdirectories

		# From https://github.com/python/cpython/pull/23424

		self.subdirs = {}

		for x in self.common_dirs:
			a_x = os.path.join(self.left, x)
			b_x = os.path.join(self.right, x)
			self.subdirs[x] = self.__class__(a_x, b_x, self.ignore, self.hide)

	_methodmap = {
			"subdirs": phase4,
			"same_files": phase3,
			"diff_files": phase3,
			"funny_files": phase3,
			"common_dirs": filecmp.dircmp.phase2,
			"common_files": filecmp.dircmp.phase2,
			"common_funny": filecmp.dircmp.phase2,
			"common": filecmp.dircmp.phase1,
			"left_only": filecmp.dircmp.phase1,
			"right_only": filecmp.dircmp.phase1,
			"left_list": filecmp.dircmp.phase0,
			"right_list": filecmp.dircmp.phase0
			}

	methodmap = _methodmap  # type: ignore


def compare_dirs(a: PathLike, b: PathLike) -> bool:
	"""
	Compare the content of two directory trees.

	.. versionadded:: 2.7.0

	:param a: The "left" directory to compare.
	:param b: The "right" directory to compare.

	:returns: :py:obj:`False` if they differ, :py:obj:`True` is they are the same.
	"""

	compared = DirComparator(a, b)

	if compared.left_only or compared.right_only or compared.diff_files or compared.funny_files:
		return False

	for subdir in compared.common_dirs:
		if not compare_dirs(os.path.join(a, subdir), os.path.join(b, subdir)):
			return False

	return True



#  !/usr/bin/env python
#
#  compat.py
r"""
Cross-version compatibility helpers.

.. versionadded :: 0.12.0

-----

Provides the following:

.. autovariable:: domdf_python_tools.compat.PYPY
	:no-value:

.. raw:: latex

	\begin{multicols}{2}

.. autovariable:: domdf_python_tools.compat.PYPY36
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY37
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY37_PLUS
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY38
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY38_PLUS
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY39
	:no-value:

.. autovariable:: domdf_python_tools.compat.PYPY39_PLUS
	:no-value:

.. raw:: latex

	\end{multicols}

.. py:data:: importlib_resources

	`importlib_resources <https://importlib-resources.readthedocs.io/en/latest/>`_ on Python 3.6;
	:mod:`importlib.resources` on Python 3.7 and later.

.. py:data:: importlib_metadata

	`importlib_metadata <https://importlib-metadata.readthedocs.io/en/latest/>`_ on Python 3.8 and earlier;
	:mod:`importlib.metadata` on Python 3.9 and later.

	.. versionadded:: 1.1.0

	.. versionchanged:: 2.5.0

		`importlib_metadata <https://importlib-metadata.readthedocs.io/en/latest/>`__ is now used
		on Python 3.8 in place of the stdlib version.
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import platform
import sys
from typing import TYPE_CHECKING, ContextManager, Optional, TypeVar

# this package
import domdf_python_tools

__all__ = [
		"importlib_resources",
		"importlib_metadata",
		"nullcontext",
		"PYPY",
		"PYPY36",
		"PYPY37",
		"PYPY37_PLUS",
		"PYPY38",
		"PYPY38_PLUS",
		"PYPY39",
		"PYPY39_PLUS",
		]

if TYPE_CHECKING:  # pragma: no cover
	# stdlib
	from contextlib import nullcontext

elif sys.version_info[:2] < (3, 7) or domdf_python_tools.__docs:  # pragma: no cover (py37+)

	_T = TypeVar("_T")

	class nullcontext(ContextManager[Optional[_T]]):
		"""
		Context manager that does no additional processing.

		Used as a stand-in for a normal context manager, when a particular
		block of code is only sometimes used with a normal context manager:

		.. code-block:: python

			cm = optional_cm if condition else nullcontext()
			with cm:
				# Perform operation, using optional_cm if condition is True

		.. versionadded:: 2.1.0

		In Python 3.7 and above the `version from the standard library`_ is used instead of this one,
		but the implementations are identical.

		.. _version from the standard library: https://docs.python.org/3/library/contextlib.html#contextlib.nullcontext

		:param enter_result: An optional value to return when entering the context.
		"""

		#  From CPython
		#  Licensed under the Python Software Foundation License Version 2.
		#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
		#  Copyright © 2000 BeOpen.com. All rights reserved.
		#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
		#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.

		def __init__(self, enter_result: Optional[_T] = None):
			self.enter_result: Optional[_T] = enter_result

		def __enter__(self) -> Optional[_T]:
			return self.enter_result

		def __exit__(self, *excinfo):
			pass

else:  # pragma: no cover (<py37)

	# stdlib
	from contextlib import nullcontext

PYPY: bool = platform.python_implementation() == "PyPy"
"""
:py:obj:`True` if running on PyPy rather than CPython.

.. versionadded:: 2.3.0
"""

PYPY36: bool = False
"""
:py:obj:`True` if running on PyPy 3.6.

.. versionadded:: 2.6.0
"""

PYPY37: bool = False
"""
:py:obj:`True` if running on PyPy 3.7.

.. versionadded:: 2.6.0
"""

PYPY38: bool = False
"""
:py:obj:`True` if running on PyPy 3.8.

.. versionadded:: 3.2.0
"""

PYPY39: bool = False
"""
:py:obj:`True` if running on PyPy 3.9.

.. versionadded:: 3.3.0
"""

PYPY37_PLUS: bool = False
"""
:py:obj:`True` if running on PyPy 3.7 or newer.

.. versionadded:: 3.3.0
"""

PYPY38_PLUS: bool = False
"""
:py:obj:`True` if running on PyPy 3.8 or newer.

.. versionadded:: 3.3.0
"""

PYPY39_PLUS: bool = False
"""
:py:obj:`True` if running on PyPy 3.9 or newer.

.. versionadded:: 3.3.0
"""

if PYPY:  # pragma: no cover
	if sys.version_info[:2] == (3, 6):
		PYPY36 = True
	elif sys.version_info[:2] == (3, 7):
		PYPY37 = True
		PYPY37_PLUS = True
	elif sys.version_info[:2] == (3, 8):
		PYPY38 = True
		PYPY38_PLUS = PYPY37_PLUS = True
	elif sys.version_info[:2] == (3, 9):
		PYPY39 = True
		PYPY39_PLUS = PYPY38_PLUS = PYPY37_PLUS = True



#!/usr/bin/env python
#
#  import_tools.py
"""
Functions for importing classes.

.. versionadded:: 0.5.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  Based on https://github.com/asottile/git-code-debt/blob/master/git_code_debt/util/discovery.py
#  Copyright (c) 2014 Anthony Sottile
#  Licensed under the MIT License
#

# stdlib
import importlib.machinery
import importlib.util
import inspect
import itertools
import pkgutil
from types import ModuleType
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, overload

# 3rd party
from typing_extensions import Literal, TypedDict

# this package
from domdf_python_tools.compat import importlib_metadata
from domdf_python_tools.paths import PathPlus, sort_paths

__all__ = [
		"discover",
		"discover_in_module",
		"discover_entry_points",
		"discover_entry_points_by_name",
		"iter_submodules"
		]


class _DiscoverKwargsType(TypedDict):
	match_func: Optional[Callable[[Any], bool]]
	exclude_side_effects: bool


@overload
def discover(
		package: ModuleType,
		match_func: Optional[Callable[[Any], bool]] = ...,
		exclude_side_effects: Literal[True] = ...,
		) -> List[Type[Any]]: ...


@overload
def discover(
		package: ModuleType,
		match_func: Optional[Callable[[Any], bool]] = ...,
		exclude_side_effects: Literal[False] = ...,
		) -> List[Any]: ...


def discover(
		package: ModuleType,
		match_func: Optional[Callable[[Any], bool]] = None,
		exclude_side_effects: bool = True,
		) -> List[Any]:
	"""
	Returns a list of objects in the given package, optionally filtered by ``match_func``.

	:param package: A Python package
	:param match_func: Function taking an object and returning :py:obj:`True` if the object is to be included in the output.
	:default match_func: :py:obj:`None`, which includes all objects.
	:param exclude_side_effects: Don't include objects that are only there because of an import side effect.

	:rtype:

	.. versionchanged:: 1.0.0  Added the ``exclude_side_effects`` parameter.
	.. latex:clearpage::
	"""

	kwargs: _DiscoverKwargsType = dict(exclude_side_effects=exclude_side_effects, match_func=match_func)

	matching_objects = discover_in_module(package, **kwargs)

	if hasattr(package, "__path__"):
		package_path = package.__path__

		for _, module_name, _ in pkgutil.walk_packages(package_path, prefix=f'{package.__name__}.'):
			module = __import__(module_name, fromlist=["__trash"], level=0)

			matching_objects.extend(discover_in_module(module, **kwargs))

	return matching_objects


def discover_in_module(
		module: ModuleType,
		match_func: Optional[Callable[[Any], bool]] = None,
		exclude_side_effects: bool = True,
		) -> List[Any]:
	"""
	Returns a list of objects in the given module, optionally filtered by ``match_func``.

	.. versionadded:: 2.6.0

	:param module: A Python module.
	:param match_func: Function taking an object and returning :py:obj:`True` if the object is to be included in the output.
	:default match_func: :py:obj:`None`, which includes all objects.
	:param exclude_side_effects: Don't include objects that are only there because of an import side effect.
	"""

	matching_objects = []

	# Check all the functions in that module
	for _, imported_object in inspect.getmembers(module, match_func):

		if exclude_side_effects:
			if not hasattr(imported_object, "__module__"):
				continue
			if imported_object.__module__ != module.__name__:
				continue

		matching_objects.append(imported_object)

	return matching_objects


#
# def import_module(filename: str):
# 	"""
# 	Import the module with the given filename.
# 	:param filename:
# 	:return:
# 	"""
#
# 	spec = importlib.util.spec_from_file_location("typing", filename)
# 	mod = importlib.util.module_from_spec(spec)
# 	spec.loader.exec_module(mod)
# 	sys.modules[mod.__name__] = mod
# 	return mod


def discover_entry_points(
		group_name: str,
		match_func: Optional[Callable[[Any], bool]] = None,
		) -> List[Any]:
	"""
	Returns a list of entry points in the given category, optionally filtered by ``match_func``.

	.. versionadded:: 1.1.0

	:param group_name: The entry point group name, e.g. ``'entry_points'``.
	:param match_func: Function taking an object and returning :py:obj:`True`
		if the object is to be included in the output.
	:default match_func: :py:obj:`None`, which includes all objects.

	:return: List of matching objects.
	"""

	return list(discover_entry_points_by_name(group_name, object_match_func=match_func).values())


def discover_entry_points_by_name(
		group_name: str,
		name_match_func: Optional[Callable[[Any], bool]] = None,
		object_match_func: Optional[Callable[[Any], bool]] = None,
		) -> Dict[str, Any]:
	"""
	Returns a mapping of entry point names to the entry points in the given category,
	optionally filtered by ``name_match_func`` and ``object_match_func``.

	.. versionadded:: 2.5.0

	:param group_name: The entry point group name, e.g. ``'entry_points'``.
	:param name_match_func: Function taking the entry point name and returning :py:obj:`True`
		if the entry point is to be included in the output.
	:default name_match_func: :py:obj:`None`, which includes all entry points.
	:param object_match_func: Function taking an object and returning :py:obj:`True`
		if the object is to be included in the output.
	:default object_match_func: :py:obj:`None`, which includes all objects.
	"""  # noqa: D400

	matching_objects = {}

	eps = itertools.chain.from_iterable(dist.entry_points for dist in importlib_metadata.distributions())

	for entry_point in eps:
		if entry_point.group != group_name:
			continue

		if name_match_func is not None and not name_match_func(entry_point.name):
			continue

		entry_point_obj = entry_point.load()

		if object_match_func is not None and not object_match_func(entry_point_obj):
			continue

		matching_objects[entry_point.name] = entry_point_obj

	return matching_objects


def iter_submodules(module: str) -> Iterator[str]:
	"""
	Returns an iterator over the names of the submodules and subpackages of the given module.

	.. versionadded:: 2.6.0

	:param module:
	"""

	spec: Optional[importlib.machinery.ModuleSpec] = importlib.util.find_spec(module)

	if spec is None or spec.origin is None:
		return

	yield module

	if spec.submodule_search_locations is None or PathPlus(spec.origin).name != "__init__.py":
		return

	for submodule_search_path in spec.submodule_search_locations:
		for item in sort_paths(*PathPlus(submodule_search_path).iterdir()):
			if item.name == "__init__.py":
				continue
			elif item.suffix == ".py":
				yield f"{module}.{item.stem}"
			elif item.name == "__pycache__":
				continue
			elif item.is_dir():
				yield from sorted(iter_submodules(f"{module}.{item.name}"))



#!/usr/bin/env python
#
#  terminal.py
"""
Useful functions for terminal-based programs.

.. versionchanged:: 2.0.0

	:func:`domdf_python_tools.terminal.get_terminal_size` was removed.
	Use :func:`shutil.get_terminal_size` instead.
"""
#
#  Copyright © 2014-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Parts of the docstrings based on the Python 3.8.2 Documentation
#  Licensed under the Python Software Foundation License Version 2.
#  Copyright © 2001-2020 Python Software Foundation. All rights reserved.
#  Copyright © 2000 BeOpen.com. All rights reserved.
#  Copyright © 1995-2000 Corporation for National Research Initiatives. All rights reserved.
#  Copyright © 1991-1995 Stichting Mathematisch Centrum. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  "Echo" based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import inspect
import os
import pprint
import textwrap
from shutil import get_terminal_size
from typing import IO, Optional

# this package
from domdf_python_tools.words import CR

__all__ = [
		"clear",
		"br",
		"interrupt",
		"overtype",
		"get_terminal_size",
		"Echo",
		]


def clear() -> None:
	"""
	Clears the display.

	Works for Windows and POSIX, but does not clear the Python Interpreter or PyCharm's Console.
	"""

	if os.name == "nt":  # pragma: no cover (!Windows)
		os.system("cls")  # nosec: B607,B605
	else:  # pragma: no cover (!Linux)
		print("\u001bc", end='')


def br() -> None:
	"""
	Prints a blank line.
	"""

	print('')


def interrupt() -> None:
	"""
	Print the key combination needed to abort the script; dynamic depending on OS.

	Useful when you have a long-running script that you might want to
	interrupt part way through.

	**Example:**

	.. code-block:: python

		>>> interrupt()
		(Press Ctrl-C to quit at any time)

	"""

	print(f"(Press Ctrl-{'C' if os.name == 'nt' else 'D'} to quit at any time)")


def overtype(
		*objects,
		sep: str = ' ',
		end: str = '',
		file: Optional[IO] = None,
		flush: bool = False,
		) -> None:
	r"""
	Print ``*objects`` to the text stream ``file``, starting with ``'\\r'``, separated by ``sep``
	and followed by ``end``.

	All non-keyword arguments are converted to strings like :class:`str` does and written to the stream,
	separated by ``sep`` and followed by ``end``.

	If no objects are given, :func:`~.overtype` will just write ``"\\r"``.

	.. TODO:: This does not currently work in the PyCharm console, at least on Windows

	:param \*objects: A list of strings or string-like objects to write to the terminal.
	:param sep: The separator between values.
	:param end: The final value to print.
	:param file: An object with a ``write(string)`` method.
		If not present or :py:obj:`None`, :py:obj:`sys.stdout` will be used.
	:no-default file:
	:param flush: If :py:obj:`True` the stream is forcibly flushed after printing.
	"""  # noqa: D400

	object0 = f"{CR}{objects[0]}"
	objects = (object0, *objects[1:])
	print(*objects, sep=sep, end=end, file=file, flush=flush)


class Echo:
	"""
	Context manager for echoing variable assignments (in CPython).

	:param indent: The indentation of the dictionary of variable assignments.
	"""

	def __init__(self, indent: str = ' ' * 2):
		self.indent = indent

		frame = inspect.currentframe()
		if frame is None:  # pragma: no cover
			raise ValueError("Unable to obtain the frame of the caller.")
		else:
			self.parent_frame = inspect.currentframe().f_back  # type: ignore  # TODO

	def __enter__(self):
		"""
		Called when entering the context manager.
		"""

		self.locals_on_entry = self.parent_frame.f_locals.copy()  # type: ignore

	def __exit__(self, *args, **kwargs):
		"""
		Called when exiting the context manager.
		"""

		new_locals = {
				k: v
				for k,
				v in self.parent_frame.f_locals.items()  # type: ignore
				if k not in self.locals_on_entry
				}

		print(textwrap.indent(pprint.pformat(new_locals), self.indent))


if __name__ == "__main__":  # pragma: no cover
	size_x, size_y = get_terminal_size()
	print("width =", size_x, "height =", size_y)



#!/usr/bin/env python
#
#  stringlist.py
"""
A list of strings that represent lines in a multiline string.

.. versionchanged:: 1.0.0

	:class:`~domdf_python_tools.typing.String` should now be imported from :mod:`domdf_python_tools.typing`.
"""
#
#  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from contextlib import contextmanager
from itertools import chain
from typing import Any, Iterable, Iterator, List, Reversible, Tuple, TypeVar, Union, cast, overload

# this package
from domdf_python_tools.doctools import prettify_docstrings
from domdf_python_tools.typing import String, SupportsIndex
from domdf_python_tools.utils import convert_indents

__all__ = ["Indent", "StringList", "DelimitedList", "_SL", "splitlines", "joinlines"]

_S = TypeVar("_S")
_SL = TypeVar("_SL", bound="StringList")


@prettify_docstrings
class Indent:
	"""
	Represents an indent, having a symbol/type and a size.

	:param size: The indent size.
	:param type: The indent character.
	"""

	def __init__(self, size: int = 0, type: str = '\t'):  # noqa: A002  # pylint: disable=redefined-builtin
		self.size = int(size)
		self.type = str(type)

	def __iter__(self) -> Iterator[Union[str, Any]]:
		"""
		Returns the size and type of the :class:`~domdf_python_tools.stringlist.Indent`.
		"""

		yield self.size
		yield self.type

	@property
	def size(self) -> int:
		"""
		The indent size.
		"""

		return self._size

	@size.setter
	def size(self, size: int) -> None:
		self._size = int(size)

	@property
	def type(self) -> str:
		"""
		The indent character.
		"""

		return self._type

	@type.setter
	def type(self, type: str) -> None:  # noqa: A002  # pylint: disable=redefined-builtin
		if not str(type):
			raise ValueError("'type' cannot an empty string.")

		self._type = str(type)

	def __str__(self) -> str:
		"""
		Returns the :class:`~domdf_python_tools.stringlist.Indent` as a string.
		"""

		return self.type * self.size

	def __repr__(self) -> str:
		"""
		Returns the string representation of the :class:`~domdf_python_tools.stringlist.Indent`.
		"""

		return f"{type(self).__name__}(size={self.size}, type={self.type!r})"

	def __eq__(self, other):
		if isinstance(other, Indent):
			return other.size == self.size and other.type == self.type
		elif isinstance(other, str):
			return str(self) == other
		elif isinstance(other, tuple):
			return tuple(self) == other
		else:
			return NotImplemented


class StringList(List[str]):
	"""
	A list of strings that represent lines in a multiline string.

	:param iterable: Content to populate the StringList with.
	:param convert_indents: Whether indents at the start of lines should be converted.
	"""

	#: The indent to insert at the beginning of new lines.
	indent: Indent

	convert_indents: bool
	"""
	Whether indents at the start of lines should be converted.

	Only applies to lines added after this is enabled/disabled.

	Can only be used when the indent is ``'\\t'`` or ``'␣'``.
	"""

	def __init__(
			self,
			iterable: Iterable[String] = (),
			convert_indents: bool = False,
			) -> None:

		if isinstance(iterable, str):
			iterable = iterable.split('\n')

		self.indent = Indent()
		self.convert_indents = convert_indents
		super().__init__([self._make_line(str(x)) for x in iterable])

	def _make_line(self, line: str) -> str:
		if not str(self.indent_type).strip(" \t") and self.convert_indents:
			if self.indent_type == '\t':
				line = convert_indents(line, tab_width=1, from_="    ", to='\t')
			else:  # pragma: no cover
				line = convert_indents(line, tab_width=1, from_='\t', to=self.indent_type)

		return f"{self.indent}{line}".rstrip()

	def append(self, line: String) -> None:
		"""
		Append a line to the end of the :class:`~domdf_python_tools.stringlist.StringList`.

		:param line:
		"""

		for inner_line in str(line).split('\n'):
			super().append(self._make_line(inner_line))

	def extend(self, iterable: Iterable[String]) -> None:
		"""
		Extend the :class:`~domdf_python_tools.stringlist.StringList` with lines from ``iterable``.

		:param iterable: An iterable of string-like objects to add to the end of the
			:class:`~domdf_python_tools.stringlist.StringList`.
		"""

		for line in iterable:
			self.append(line)

	def copy(self: _SL) -> _SL:
		"""
		Returns a shallow copy of the :class:`~domdf_python_tools.stringlist.StringList`.

		Equivalent to ``a[:]``.

		:rtype: :class:`~domdf_python_tools.stringlist.StringList`
		"""

		return self.__class__(super().copy())

	def count_blanklines(self) -> int:
		"""
		Returns a count of the blank lines in the :class:`~domdf_python_tools.stringlist.StringList`.

		.. versionadded:: 0.7.1
		"""

		return self.count('')

	def insert(self, index: SupportsIndex, line: String) -> None:
		"""
		Insert a line into the :class:`~domdf_python_tools.stringlist.StringList` at the given position.

		:param index:
		:param line:

		.. versionchanged:: 3.2.0  Changed :class:`int` in the type annotation to :protocol:`~.SupportsIndex`.
		"""

		lines: List[str]

		index = index.__index__()

		if index < 0 or index > len(self):
			lines = str(line).split('\n')
		else:
			lines = cast(list, reversed(str(line).split('\n')))

		for inner_line in lines:
			super().insert(index, self._make_line(inner_line))

	@overload
	def __setitem__(self, index: SupportsIndex, line: String) -> None: ...

	@overload
	def __setitem__(self, index: slice, line: Iterable[String]) -> None: ...

	def __setitem__(self, index: Union[SupportsIndex, slice], line: Union[String, Iterable[String]]):
		"""
		Replaces the given line with new content.

		If the new content consists of multiple lines subsequent content in the
		:class:`~domdf_python_tools.stringlist.StringList` will be shifted down.

		:param index:
		:param line:

		.. versionchanged:: 3.2.0  Changed :class:`int` in the type annotation to :protocol:`~.SupportsIndex`.
		"""

		if isinstance(index, slice):
			line = cast(Iterable[String], line)

			if not isinstance(line, Reversible):
				line = tuple(line)

			for lline, idx in zip(
				reversed(line),
				reversed(range(index.start or 0, index.stop + 1, index.step or 1)),
				):
				self[idx] = lline
		else:
			line = cast(String, line)
			index = index.__index__()

			if self and index < len(self):
				self.pop(index)
			if index < 0:
				index = len(self) + index + 1

			self.insert(index, line)

	@overload
	def __getitem__(self, index: SupportsIndex) -> str: ...

	@overload
	def __getitem__(self: _SL, index: slice) -> _SL: ...

	def __getitem__(self: _SL, index: Union[SupportsIndex, slice]) -> Union[str, _SL]:
		r"""
		Returns the line with the given index.

		:param index:

		:rtype: :py:obj:`~typing.Union`\[:class:`str`, :class:`~domdf_python_tools.stringlist.StringList`\]

		.. versionchanged:: 1.8.0

			Now returns a :class:`~domdf_python_tools.stringlist.StringList` when ``index`` is a :class:`slice`.

		.. versionchanged:: 3.2.0  Changed :class:`int` in the type annotation to :protocol:`~.SupportsIndex`.
		"""

		if isinstance(index, slice):
			return self.__class__(super().__getitem__(index))
		else:
			return super().__getitem__(index)

	def blankline(self, ensure_single: bool = False):
		"""
		Append a blank line to the end of the :class:`~domdf_python_tools.stringlist.StringList`.

		:param ensure_single: Ensure only a single blank line exists after the previous line of text.
		"""

		if ensure_single:
			while self and not self[-1]:
				self.pop(-1)

		self.append('')

	def set_indent_size(self, size: int = 0):
		"""
		Sets the size of the indent to insert at the beginning of new lines.

		:param size: The indent size to use for new lines.
		"""

		self.indent.size = int(size)

	def set_indent_type(self, indent_type: str = '\t'):
		"""
		Sets the type of the indent to insert at the beginning of new lines.

		:param indent_type: The type of indent to use for new lines.
		"""

		self.indent.type = str(indent_type)

	def set_indent(self, indent: Union[String, Indent], size: int = 0):
		"""
		Sets the indent to insert at the beginning of new lines.

		:param indent: The :class:`~.Indent` to use for new lines, or the indent type.
		:param size: If ``indent`` is an indent type, the indent size to use for new lines.
		"""

		if isinstance(indent, Indent):
			if size:
				raise TypeError("'size' argument cannot be used when providing an 'Indent' object.")

			self.indent = indent
		else:
			self.indent = Indent(int(size), str(indent))

	@property
	def indent_size(self) -> int:
		"""
		The current indent size.
		"""

		return int(self.indent.size)

	@indent_size.setter
	def indent_size(self, size: int) -> None:
		"""
		Sets the indent size.
		"""

		self.indent.size = int(size)

	@property
	def indent_type(self) -> str:
		"""
		The current indent type.
		"""

		return str(self.indent.type)

	@indent_type.setter
	def indent_type(self, type: str) -> None:  # noqa: A002  # pylint: disable=redefined-builtin
		"""
		Sets the indent type.
		"""

		self.indent.type = str(type)

	def __str__(self) -> str:
		"""
		Returns the :class:`~domdf_python_tools.stringlist.StringList` as a string.
		"""

		return '\n'.join(self)

	def __bytes__(self) -> bytes:
		"""
		Returns the :class:`~domdf_python_tools.stringlist.StringList` as bytes.

		.. versionadded:: 2.1.0
		"""

		return str(self).encode("UTF-8")

	def __eq__(self, other) -> bool:
		"""
		Returns whether the other object is equal to this :class:`~domdf_python_tools.stringlist.StringList`.
		"""

		if isinstance(other, str):
			return str(self) == other
		else:
			return super().__eq__(other)

	@contextmanager
	def with_indent(self, indent: Union[String, Indent], size: int = 0):
		"""
		Context manager to temporarily use a different indent.

		.. code-block:: python

			>>> sl = StringList()
			>>> with sl.with_indent("    ", 1):
			...     sl.append("Hello World")

		:param indent: The :class:`~.Indent` to use within the ``with`` block, or the indent type.
		:param size: If ``indent`` is an indent type, the indent size to use within the ``with`` block.
		"""

		original_indent: Tuple[int, str] = tuple(self.indent)  # type: ignore

		try:
			self.set_indent(indent, size)
			yield
		finally:
			self.indent = Indent(*original_indent)

	@contextmanager
	def with_indent_size(self, size: int = 0):
		"""
		Context manager to temporarily use a different indent size.

		.. code-block:: python

			>>> sl = StringList()
			>>> with sl.with_indent_size(1):
			...     sl.append("Hello World")

		:param size: The indent size to use within the ``with`` block.
		"""

		original_indent_size = self.indent_size

		try:
			self.indent_size = size
			yield
		finally:
			self.indent_size = original_indent_size

	@contextmanager
	def with_indent_type(self, indent_type: str = '\t'):
		"""
		Context manager to temporarily use a different indent type.

		.. code-block:: python

			>>> sl = StringList()
			>>> with sl.with_indent_type("    "):
			...     sl.append("Hello World")

		:param indent_type: The type of indent to use within the ``with`` block.
		"""

		original_indent_type = self.indent_type

		try:
			self.indent_type = indent_type
			yield
		finally:
			self.indent_type = original_indent_type

	def splitlines(self, keepends: bool = False) -> List[str]:
		"""
		Analagous to :meth:`str.splitlines`.

		.. versionadded:: 3.8.0
		"""

		if keepends:
			return [line + '\n' for line in self]
		else:
			return self


class DelimitedList(List[_S]):
	"""
	Subclass of :class:`list` that supports custom delimiters in format strings.

	**Example:**

	.. code-block:: python

		>>> l = DelimitedList([1, 2, 3, 4, 5])
		>>> format(l, ", ")
		'1, 2, 3, 4, 5'
		>>> f"Numbers: {l:, }"
		'Numbers: 1, 2, 3, 4, 5'

	.. autoclasssumm:: DelimitedList
		:autosummary-sections: ;;

	.. versionadded:: 1.1.0
	"""

	def __format__(self, format_spec: str) -> str:
		return format_spec.join([str(x) for x in self])  # pylint: disable=not-an-iterable


def splitlines(string: str) -> List[Tuple[str, str]]:
	"""
	Split ``string`` into a list of two-element tuples,
	containing the line content and the newline character(s), if any.

	.. versionadded:: 3.2.0

	:param string:

	:rtype:

	.. seealso:: :meth:`str.splitlines` and :func:`~.stringlist.joinlines`
	"""  # noqa: D400

	# Translated and adapted from https://github.com/python/cpython/blob/main/Objects/stringlib/split.h

	str_len: int = len(string)
	i: int = 0
	j: int = 0
	eol: int
	the_list: List[Tuple[str, str]] = []

	while i < str_len:

		# Find a line and append it
		while i < str_len and string[i] not in "\n\r":
			i += 1

		# Skip the line break reading CRLF as one line break
		eol = i
		if i < str_len:
			if (string[i] == '\r') and (i + 1 < str_len) and (string[i + 1] == '\n'):
				i += 2
			else:
				i += 1

		if j == 0 and eol == str_len and type(string) is str:  # pylint: disable=unidiomatic-typecheck
			# No whitespace in string, so just use it as the_list[0]
			the_list.append((string, ''))
			break

		the_list.append((string[j:eol], string[eol:i]))
		j = i

	return the_list


def joinlines(lines: List[Tuple[str, str]]) -> str:
	"""
	Given a list of two-element tuples, each containing a line and a newline character (or empty string),
	return a single string.

	.. versionadded:: 3.2.0

	:param lines:

	:rtype:

	.. seealso:: :func:`~.stringlist.splitlines`
	"""  # noqa: D400

	return ''.join(chain.from_iterable(lines))
